"""
GitLab API:
https://docs.gitlab.com/ee/api/personal_access_tokens.html
https://docs.gitlab.com/ee/api/users.html#create-a-personal-access-token
"""

import pytest
import responses

from gitlab.v4.objects import PersonalAccessToken

user_id = 1
token_id = 1
token_name = "Test Token"

token_url = "http://localhost/api/v4/personal_access_tokens"
single_token_url = f"{token_url}/{token_id}"
self_token_url = f"{token_url}/self"
user_token_url = f"http://localhost/api/v4/users/{user_id}/personal_access_tokens"

content = {
    "id": token_id,
    "name": token_name,
    "revoked": False,
    "created_at": "2020-07-23T14:31:47.729Z",
    "scopes": ["api"],
    "active": True,
    "user_id": user_id,
    "expires_at": None,
}


@pytest.fixture
def resp_create_user_personal_access_token():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=user_token_url,
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_personal_access_tokens():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=token_url,
            json=[content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_personal_access_token():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=single_token_url,
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_personal_access_token_self():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=self_token_url,
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_delete_personal_access_token():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url=single_token_url,
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_rotate_personal_access_token(token_content):
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/personal_access_tokens/1/rotate",
            json=token_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_create_personal_access_token(gl, resp_create_user_personal_access_token):
    user = gl.users.get(1, lazy=True)
    access_token = user.personal_access_tokens.create(
        {"name": token_name, "scopes": "api"}
    )
    assert access_token.revoked is False
    assert access_token.name == token_name


def test_list_personal_access_tokens(gl, resp_list_personal_access_tokens):
    access_tokens = gl.personal_access_tokens.list()
    assert len(access_tokens) == 1
    assert access_tokens[0].revoked is False
    assert access_tokens[0].name == token_name


def test_list_personal_access_tokens_filter(gl, resp_list_personal_access_tokens):
    access_tokens = gl.personal_access_tokens.list(user_id=user_id)
    assert len(access_tokens) == 1
    assert access_tokens[0].revoked is False
    assert access_tokens[0].user_id == user_id


def test_get_personal_access_token(gl, resp_get_personal_access_token):
    access_token = gl.personal_access_tokens.get(token_id)

    assert access_token.revoked is False
    assert access_token.user_id == user_id


def test_get_personal_access_token_self(gl, resp_get_personal_access_token_self):
    access_token = gl.personal_access_tokens.get("self")

    assert access_token.revoked is False
    assert access_token.user_id == user_id


def test_delete_personal_access_token(gl, resp_delete_personal_access_token):
    access_token = gl.personal_access_tokens.get(token_id, lazy=True)
    access_token.delete()


def test_revoke_personal_access_token_by_id(gl, resp_delete_personal_access_token):
    gl.personal_access_tokens.delete(token_id)


def test_rotate_project_access_token(gl, resp_rotate_personal_access_token):
    access_token = gl.personal_access_tokens.get(1, lazy=True)
    access_token.rotate()
    assert isinstance(access_token, PersonalAccessToken)
    assert access_token.token == "s3cr3t"



"""
GitLab API: https://docs.gitlab.com/ee/api/project_container_registry_protection_rules.html
"""

import pytest
import responses

from gitlab.v4.objects import ProjectRegistryProtectionRule

protected_registry_content = {
    "id": 1,
    "project_id": 7,
    "repository_path_pattern": "test/image",
    "minimum_access_level_for_push": "maintainer",
    "minimum_access_level_for_delete": "maintainer",
}


@pytest.fixture
def resp_list_protected_registries():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/registry/protection/rules",
            json=[protected_registry_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_protected_registry():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/registry/protection/rules",
            json=protected_registry_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_update_protected_registry():
    updated_content = protected_registry_content.copy()
    updated_content["repository_path_pattern"] = "abc*"

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PATCH,
            url="http://localhost/api/v4/projects/1/registry/protection/rules/1",
            json=updated_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_project_protected_registries(project, resp_list_protected_registries):
    protected_registry = project.registry_protection_rules.list()[0]
    assert isinstance(protected_registry, ProjectRegistryProtectionRule)
    assert protected_registry.repository_path_pattern == "test/image"


def test_create_project_protected_registry(project, resp_create_protected_registry):
    protected_registry = project.registry_protection_rules.create(
        {
            "repository_path_pattern": "test/image",
            "minimum_access_level_for_push": "maintainer",
        }
    )
    assert isinstance(protected_registry, ProjectRegistryProtectionRule)
    assert protected_registry.repository_path_pattern == "test/image"


def test_update_project_protected_registry(project, resp_update_protected_registry):
    updated = project.registry_protection_rules.update(
        1, {"repository_path_pattern": "abc*"}
    )
    assert updated["repository_path_pattern"] == "abc*"



"""
Ensure objects defined in gitlab.v4.objects are imported in
`gitlab/v4/objects/__init__.py`

"""

import pkgutil
from typing import Set

import gitlab.exceptions
import gitlab.v4.objects


def test_all_exceptions_imports_are_exported() -> None:
    assert gitlab.exceptions.__all__ == sorted(
        [
            name
            for name in dir(gitlab.exceptions)
            if name.endswith("Error") and not name.startswith("_")
        ]
    )


def test_all_v4_objects_are_imported() -> None:
    assert len(gitlab.v4.objects.__path__) == 1

    init_files: Set[str] = set()
    with open(gitlab.v4.objects.__file__, "r", encoding="utf-8") as in_file:
        for line in in_file.readlines():
            if line.startswith("from ."):
                init_files.add(line.rstrip())

    object_files = set()
    for module in pkgutil.iter_modules(gitlab.v4.objects.__path__):
        object_files.add(f"from .{module.name} import *")

    missing_in_init = object_files - init_files
    error_message = (
        f"\nThe file {gitlab.v4.objects.__file__!r} is missing the following imports:"
    )
    for missing in sorted(missing_in_init):
        error_message += f"\n    {missing}"

    assert not missing_in_init, error_message



"""
GitLab API: https://docs.gitlab.com/ee/api/jobs.html
"""

from functools import partial

import pytest
import responses

from gitlab.v4.objects import ProjectJob

failed_job_content = {
    "commit": {
        "author_email": "admin@example.com",
        "author_name": "Administrator",
    },
    "coverage": None,
    "allow_failure": False,
    "created_at": "2015-12-24T15:51:21.880Z",
    "started_at": "2015-12-24T17:54:30.733Z",
    "finished_at": "2015-12-24T17:54:31.198Z",
    "duration": 0.465,
    "queued_duration": 0.010,
    "artifacts_expire_at": "2016-01-23T17:54:31.198Z",
    "tag_list": ["docker runner", "macos-10.15"],
    "id": 1,
    "name": "rubocop",
    "pipeline": {
        "id": 1,
        "project_id": 1,
    },
    "ref": "main",
    "artifacts": [],
    "runner": None,
    "stage": "test",
    "status": "failed",
    "tag": False,
    "web_url": "https://example.com/foo/bar/-/jobs/1",
    "user": {"id": 1},
}

success_job_content = {
    **failed_job_content,
    "status": "success",
    "id": failed_job_content["id"] + 1,
}


@pytest.fixture
def resp_get_job():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/jobs/1",
            json=failed_job_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_cancel_job():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/jobs/1/cancel",
            json=failed_job_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_retry_job():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/jobs/1/retry",
            json=failed_job_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_list_job():
    urls = [
        "http://localhost/api/v4/projects/1/jobs",
        "http://localhost/api/v4/projects/1/pipelines/1/jobs",
    ]
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        register_endpoint = partial(
            rsps.add,
            method=responses.GET,
            content_type="application/json",
            status=200,
        )
        for url in urls:
            register_endpoint(
                url=url,
                json=[failed_job_content],
                match=[responses.matchers.query_param_matcher({"scope[]": "failed"})],
            )
            register_endpoint(
                url=url,
                json=[success_job_content],
                match=[responses.matchers.query_param_matcher({"scope[]": "success"})],
            )
            register_endpoint(
                url=url,
                json=[success_job_content, failed_job_content],
                match=[
                    responses.matchers.query_string_matcher(
                        "scope[]=success&scope[]failed"
                    )
                ],
            )
            register_endpoint(
                url=url,
                json=[success_job_content, failed_job_content],
            )
        yield rsps


def test_get_project_job(project, resp_get_job):
    job = project.jobs.get(1)
    assert isinstance(job, ProjectJob)
    assert job.ref == "main"


def test_cancel_project_job(project, resp_cancel_job):
    job = project.jobs.get(1, lazy=True)

    output = job.cancel()
    assert output["ref"] == "main"


def test_retry_project_job(project, resp_retry_job):
    job = project.jobs.get(1, lazy=True)

    output = job.retry()
    assert output["ref"] == "main"


def test_list_project_job(project, resp_list_job):
    failed_jobs = project.jobs.list(scope="failed")
    success_jobs = project.jobs.list(scope="success")
    failed_and_success_jobs = project.jobs.list(scope=["failed", "success"])
    pipeline_lazy = project.pipelines.get(1, lazy=True)
    pjobs_failed = pipeline_lazy.jobs.list(scope="failed")
    pjobs_success = pipeline_lazy.jobs.list(scope="success")
    pjobs_failed_and_success = pipeline_lazy.jobs.list(scope=["failed", "success"])

    prepared_urls = [c.request.url for c in resp_list_job.calls]

    # Both pipelines and pipelines/jobs should behave the same way
    # When `scope` is scalar, one can use scope=value or scope[]=value
    assert set(failed_and_success_jobs) == set(failed_jobs + success_jobs)
    assert set(pjobs_failed_and_success) == set(pjobs_failed + pjobs_success)
    assert prepared_urls == [
        "http://localhost/api/v4/projects/1/jobs?scope%5B%5D=failed",
        "http://localhost/api/v4/projects/1/jobs?scope%5B%5D=success",
        "http://localhost/api/v4/projects/1/jobs?scope%5B%5D=failed&scope%5B%5D=success",
        "http://localhost/api/v4/projects/1/pipelines/1/jobs?scope%5B%5D=failed",
        "http://localhost/api/v4/projects/1/pipelines/1/jobs?scope%5B%5D=success",
        "http://localhost/api/v4/projects/1/pipelines/1/jobs?scope%5B%5D=failed&scope%5B%5D=success",
    ]



import pytest

pytestmark = pytest.mark.gitlab_premium


def test_epics(group):
    epic = group.epics.create({"title": "Test epic"})
    epic.title = "Fixed title"
    epic.labels = ["label1", "label2"]
    epic.save()

    epic = group.epics.get(epic.iid)
    assert epic.title == "Fixed title"
    assert epic.labels == ["label1", "label2"]
    assert group.epics.list()


@pytest.mark.xfail(reason="404 on issue.id")
def test_epic_issues(epic, issue):
    assert not epic.issues.list()

    epic_issue = epic.issues.create({"issue_id": issue.id})
    assert epic.issues.list()

    epic_issue.delete()


def test_epic_notes(epic):
    assert not epic.notes.list()

    epic.notes.create({"body": "Test note"})
    assert epic.notes.list()



import logging
import time
from typing import Optional, TYPE_CHECKING

import pytest

import gitlab
import gitlab.base
import gitlab.exceptions

SLEEP_INTERVAL = 0.5
TIMEOUT = 60  # seconds before timeout will occur
MAX_ITERATIONS = int(TIMEOUT / SLEEP_INTERVAL)


def get_gitlab_plan(gl: gitlab.Gitlab) -> Optional[str]:
    """Determine the license available on the GitLab instance"""
    try:
        license = gl.get_license()
    except gitlab.exceptions.GitlabLicenseError:
        # Without a license we assume only Free features are available
        return None

    if TYPE_CHECKING:
        assert isinstance(license["plan"], str)
    return license["plan"]


def safe_delete(object: gitlab.base.RESTObject) -> None:
    """Ensure the object specified can not be retrieved. If object still exists after
    timeout period, fail the test"""
    manager = object.manager
    for index in range(MAX_ITERATIONS):
        try:
            object = manager.get(object.get_id())  # type: ignore[attr-defined]
        except gitlab.exceptions.GitlabGetError:
            return

        if index:
            logging.info(f"Attempt {index + 1} to delete {object!r}.")
        try:
            if isinstance(object, gitlab.v4.objects.User):
                # You can't use this option if the selected user is the sole owner of any groups
                # Use `hard_delete=True` or a 'Ghost User' may be created.
                # https://docs.gitlab.com/ee/api/users.html#user-deletion
                object.delete(hard_delete=True)
                if index > 1:
                    # If User is the sole owner of any group it won't be deleted,
                    # which combined with parents group never immediately deleting in GL 16
                    # we shouldn't cause test to fail if it still exists
                    return
            elif isinstance(object, gitlab.v4.objects.Project):
                # Immediately delete rather than waiting for at least 1day
                # https://docs.gitlab.com/ee/api/projects.html#delete-project
                object.delete(permanently_remove=True)
                pass
            else:
                # We only attempt to delete parent groups to prevent dangling sub-groups
                # However parent groups can only be deleted on a delay in Gl 16
                # https://docs.gitlab.com/ee/api/groups.html#remove-group
                object.delete()
        except gitlab.exceptions.GitlabDeleteError:
            logging.info(f"{object!r} already deleted or scheduled for deletion.")
            if isinstance(object, gitlab.v4.objects.Group):
                # Parent groups can never be immediately deleted in GL 16,
                # so don't cause test to fail if it still exists
                return
            pass

        time.sleep(SLEEP_INTERVAL)
    pytest.fail(f"{object!r} was not deleted")



import pathlib

import pytest
import requests
import responses
from requests import PreparedRequest

from gitlab import Gitlab
from gitlab._backends import JobTokenAuth, OAuthTokenAuth, PrivateTokenAuth
from gitlab.config import GitlabConfigParser


@pytest.fixture
def netrc(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path):
    netrc_file = tmp_path / ".netrc"
    netrc_file.write_text("machine localhost login test password test")
    monkeypatch.setenv("NETRC", str(netrc_file))


def test_invalid_auth_args():
    with pytest.raises(ValueError):
        Gitlab(
            "http://localhost",
            api_version="4",
            private_token="private_token",
            oauth_token="bearer",
        )
    with pytest.raises(ValueError):
        Gitlab(
            "http://localhost",
            api_version="4",
            oauth_token="bearer",
            http_username="foo",
            http_password="bar",
        )
    with pytest.raises(ValueError):
        Gitlab(
            "http://localhost",
            api_version="4",
            private_token="private_token",
            http_password="bar",
        )
    with pytest.raises(ValueError):
        Gitlab(
            "http://localhost",
            api_version="4",
            private_token="private_token",
            http_username="foo",
        )


def test_private_token_auth():
    gl = Gitlab("http://localhost", private_token="private_token", api_version="4")
    p = PreparedRequest()
    p.prepare(url=gl.url, auth=gl._auth)
    assert gl.private_token == "private_token"
    assert gl.oauth_token is None
    assert gl.job_token is None
    assert isinstance(gl._auth, PrivateTokenAuth)
    assert gl._auth.token == "private_token"
    assert p.headers["PRIVATE-TOKEN"] == "private_token"
    assert "JOB-TOKEN" not in p.headers
    assert "Authorization" not in p.headers


def test_oauth_token_auth():
    gl = Gitlab("http://localhost", oauth_token="oauth_token", api_version="4")
    p = PreparedRequest()
    p.prepare(url=gl.url, auth=gl._auth)
    assert gl.private_token is None
    assert gl.oauth_token == "oauth_token"
    assert gl.job_token is None
    assert isinstance(gl._auth, OAuthTokenAuth)
    assert gl._auth.token == "oauth_token"
    assert p.headers["Authorization"] == "Bearer oauth_token"
    assert "PRIVATE-TOKEN" not in p.headers
    assert "JOB-TOKEN" not in p.headers


def test_job_token_auth():
    gl = Gitlab("http://localhost", job_token="CI_JOB_TOKEN", api_version="4")
    p = PreparedRequest()
    p.prepare(url=gl.url, auth=gl._auth)
    assert gl.private_token is None
    assert gl.oauth_token is None
    assert gl.job_token == "CI_JOB_TOKEN"
    assert isinstance(gl._auth, JobTokenAuth)
    assert gl._auth.token == "CI_JOB_TOKEN"
    assert p.headers["JOB-TOKEN"] == "CI_JOB_TOKEN"
    assert "PRIVATE-TOKEN" not in p.headers
    assert "Authorization" not in p.headers


def test_http_auth():
    gl = Gitlab(
        "http://localhost",
        http_username="foo",
        http_password="bar",
        api_version="4",
    )
    p = PreparedRequest()
    p.prepare(url=gl.url, auth=gl._auth)
    assert gl.private_token is None
    assert gl.oauth_token is None
    assert gl.job_token is None
    assert isinstance(gl._auth, requests.auth.HTTPBasicAuth)
    assert gl._auth.username == "foo"
    assert gl._auth.password == "bar"
    assert p.headers["Authorization"] == "Basic Zm9vOmJhcg=="
    assert "PRIVATE-TOKEN" not in p.headers
    assert "JOB-TOKEN" not in p.headers


@responses.activate
def test_with_no_auth_uses_netrc_file(netrc):
    responses.get(
        url="http://localhost/api/v4/test",
        match=[
            responses.matchers.header_matcher({"Authorization": "Basic dGVzdDp0ZXN0"})
        ],
    )

    gl = Gitlab("http://localhost")
    gl.http_get("/test")


@responses.activate
def test_with_auth_ignores_netrc_file(netrc):
    responses.get(
        url="http://localhost/api/v4/test",
        match=[responses.matchers.header_matcher({"Authorization": "Bearer test"})],
    )

    gl = Gitlab("http://localhost", oauth_token="test")
    gl.http_get("/test")


@pytest.mark.parametrize(
    "options,config,expected_private_token,expected_oauth_token,expected_job_token",
    [
        (
            {
                "private_token": "options-private-token",
                "oauth_token": "options-oauth-token",
                "job_token": "options-job-token",
            },
            {
                "private_token": "config-private-token",
                "oauth_token": "config-oauth-token",
                "job_token": "config-job-token",
            },
            "options-private-token",
            None,
            None,
        ),
        (
            {
                "private_token": None,
                "oauth_token": "options-oauth-token",
                "job_token": "options-job-token",
            },
            {
                "private_token": "config-private-token",
                "oauth_token": "config-oauth-token",
                "job_token": "config-job-token",
            },
            "config-private-token",
            None,
            None,
        ),
        (
            {
                "private_token": None,
                "oauth_token": None,
                "job_token": "options-job-token",
            },
            {
                "private_token": "config-private-token",
                "oauth_token": "config-oauth-token",
                "job_token": "config-job-token",
            },
            "config-private-token",
            None,
            None,
        ),
        (
            {
                "private_token": None,
                "oauth_token": None,
                "job_token": None,
            },
            {
                "private_token": "config-private-token",
                "oauth_token": "config-oauth-token",
                "job_token": "config-job-token",
            },
            "config-private-token",
            None,
            None,
        ),
        (
            {
                "private_token": None,
                "oauth_token": None,
                "job_token": None,
            },
            {
                "private_token": None,
                "oauth_token": "config-oauth-token",
                "job_token": "config-job-token",
            },
            None,
            "config-oauth-token",
            None,
        ),
        (
            {
                "private_token": None,
                "oauth_token": None,
                "job_token": None,
            },
            {
                "private_token": None,
                "oauth_token": None,
                "job_token": "config-job-token",
            },
            None,
            None,
            "config-job-token",
        ),
    ],
)
def test_merge_auth(
    options,
    config,
    expected_private_token,
    expected_oauth_token,
    expected_job_token,
):
    cp = GitlabConfigParser()
    cp.private_token = config["private_token"]
    cp.oauth_token = config["oauth_token"]
    cp.job_token = config["job_token"]

    private_token, oauth_token, job_token = Gitlab._merge_auth(options, cp)
    assert private_token == expected_private_token
    assert oauth_token == expected_oauth_token
    assert job_token == expected_job_token



"""
GitLab API:
https://docs.gitlab.com/ce/api/packages.html
https://docs.gitlab.com/ee/user/packages/generic_packages
"""

from collections.abc import Iterator

import pytest

from gitlab import Gitlab
from gitlab.v4.objects import GenericPackage, Project, ProjectPackageProtectionRule

package_name = "hello-world"
package_version = "v1.0.0"
file_name = "hello.tar.gz"
file_name2 = "hello2.tar.gz"
file_content = "package content"


@pytest.fixture(scope="module", autouse=True)
def protected_package_feature(gl: Gitlab):
    gl.features.set(name="packages_protected_packages", value=True)


def test_list_project_packages(project):
    packages = project.packages.list()
    assert isinstance(packages, list)


def test_list_group_packages(group):
    packages = group.packages.list()
    assert isinstance(packages, list)


def test_upload_generic_package(tmp_path, project):
    path = tmp_path / file_name
    path.write_text(file_content)
    package = project.generic_packages.upload(
        package_name=package_name,
        package_version=package_version,
        file_name=file_name,
        path=path,
    )

    assert isinstance(package, GenericPackage)
    assert package.message == "201 Created"


def test_upload_generic_package_as_bytes(tmp_path, project):
    path = tmp_path / file_name

    path.write_text(file_content)

    package = project.generic_packages.upload(
        package_name=package_name,
        package_version=package_version,
        file_name=file_name,
        data=path.read_bytes(),
    )

    assert isinstance(package, GenericPackage)
    assert package.message == "201 Created"


def test_upload_generic_package_as_file(tmp_path, project):
    path = tmp_path / file_name

    path.write_text(file_content)

    package = project.generic_packages.upload(
        package_name=package_name,
        package_version=package_version,
        file_name=file_name,
        data=path.open(mode="rb"),
    )

    assert isinstance(package, GenericPackage)
    assert package.message == "201 Created"


def test_upload_generic_package_select(tmp_path, project):
    path = tmp_path / file_name2
    path.write_text(file_content)
    package = project.generic_packages.upload(
        package_name=package_name,
        package_version=package_version,
        file_name=file_name2,
        path=path,
        select="package_file",
    )

    assert isinstance(package, GenericPackage)
    assert package.file_name == file_name2
    assert package.size == path.stat().st_size


def test_download_generic_package(project):
    package = project.generic_packages.download(
        package_name=package_name,
        package_version=package_version,
        file_name=file_name,
    )

    assert isinstance(package, bytes)
    assert package.decode("utf-8") == file_content


def test_stream_generic_package(project):
    bytes_iterator = project.generic_packages.download(
        package_name=package_name,
        package_version=package_version,
        file_name=file_name,
        iterator=True,
    )

    assert isinstance(bytes_iterator, Iterator)

    package = bytes()
    for chunk in bytes_iterator:
        package += chunk

    assert isinstance(package, bytes)
    assert package.decode("utf-8") == file_content


def test_download_generic_package_to_file(tmp_path, project):
    path = tmp_path / file_name

    with open(path, "wb") as f:
        project.generic_packages.download(
            package_name=package_name,
            package_version=package_version,
            file_name=file_name,
            streamed=True,
            action=f.write,
        )

    with open(path, "r") as f:
        assert f.read() == file_content


def test_stream_generic_package_to_file(tmp_path, project):
    path = tmp_path / file_name

    bytes_iterator = project.generic_packages.download(
        package_name=package_name,
        package_version=package_version,
        file_name=file_name,
        iterator=True,
    )

    with open(path, "wb") as f:
        for chunk in bytes_iterator:
            f.write(chunk)

    with open(path, "r") as f:
        assert f.read() == file_content


def test_list_project_protected_packages(project: Project):
    rules = project.package_protection_rules.list()
    assert isinstance(rules, list)


@pytest.mark.skip(reason="Not released yet")
def test_create_project_protected_packages(project: Project):
    protected_package = project.package_protection_rules.create(
        {
            "package_name_pattern": "v*",
            "package_type": "npm",
            "minimum_access_level_for_push": "maintainer",
        }
    )
    assert isinstance(protected_package, ProjectPackageProtectionRule)
    assert protected_package.package_type == "npm"

    protected_package.minimum_access_level_for_push = "owner"
    protected_package.save()

    protected_package.delete()



"""
GitLab API:
https://docs.gitlab.com/ce/api/topics.html
"""


def test_topics(gl, gitlab_version):
    assert not gl.topics.list()

    create_dict = {"name": "my-topic", "description": "My Topic"}
    if gitlab_version.major >= 15:
        create_dict["title"] = "my topic title"
    topic = gl.topics.create(create_dict)
    assert topic.name == "my-topic"

    if gitlab_version.major >= 15:
        assert topic.title == "my topic title"

    assert gl.topics.list()

    topic.description = "My Updated Topic"
    topic.save()
    updated_topic = gl.topics.get(topic.id)
    assert updated_topic.description == topic.description

    create_dict = {"name": "my-second-topic", "description": "My Second Topic"}
    if gitlab_version.major >= 15:
        create_dict["title"] = "my second topic title"
    topic2 = gl.topics.create(create_dict)
    merged_topic = gl.topics.merge(topic.id, topic2.id)
    assert merged_topic["id"] == topic2.id

    topic2.delete()



"""
GitLab API: https://docs.gitlab.com/ee/api/project_job_token_scopes.html
"""

import pytest
import responses

from gitlab.v4.objects import ProjectJobTokenScope
from gitlab.v4.objects.job_token_scope import (
    AllowlistGroupManager,
    AllowlistProjectManager,
)

job_token_scope_content = {
    "inbound_enabled": True,
    "outbound_enabled": False,
}

project_allowlist_content = [
    {
        "id": 4,
        "description": "",
        "name": "Diaspora Client",
        "name_with_namespace": "Diaspora / Diaspora Client",
        "path": "diaspora-client",
        "path_with_namespace": "diaspora/diaspora-client",
        "created_at": "2013-09-30T13:46:02Z",
        "default_branch": "main",
        "tag_list": ["example", "disapora client"],
        "topics": ["example", "disapora client"],
        "ssh_url_to_repo": "git@gitlab.example.com:diaspora/diaspora-client.git",
        "http_url_to_repo": "https://gitlab.example.com/diaspora/diaspora-client.git",
        "web_url": "https://gitlab.example.com/diaspora/diaspora-client",
        "avatar_url": "https://gitlab.example.com/uploads/project/avatar/4/uploads/avatar.png",
        "star_count": 0,
        "last_activity_at": "2013-09-30T13:46:02Z",
        "namespace": {
            "id": 2,
            "name": "Diaspora",
            "path": "diaspora",
            "kind": "group",
            "full_path": "diaspora",
            "parent_id": "",
            "avatar_url": "",
            "web_url": "https://gitlab.example.com/diaspora",
        },
    }
]

project_allowlist_created_content = {
    "target_project_id": 2,
    "project_id": 1,
}

groups_allowlist_content = [
    {
        "id": 4,
        "web_url": "https://gitlab.example.com/groups/diaspora/diaspora-group",
        "name": "namegroup",
    }
]

group_allowlist_created_content = {
    "target_group_id": 4,
    "project_id": 1,
}


@pytest.fixture
def resp_get_job_token_scope():
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/job_token_scope",
            json=job_token_scope_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_allowlist():
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/job_token_scope/allowlist",
            json=project_allowlist_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_add_to_allowlist():
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/job_token_scope/allowlist",
            json=project_allowlist_created_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_groups_allowlist():
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/job_token_scope/groups_allowlist",
            json=groups_allowlist_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_add_to_groups_allowlist():
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/job_token_scope/groups_allowlist",
            json=group_allowlist_created_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_patch_job_token_scope():
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.PATCH,
            url="http://localhost/api/v4/projects/1/job_token_scope",
            status=204,
            match=[responses.matchers.json_params_matcher({"enabled": False})],
        )
        yield rsps


@pytest.fixture
def job_token_scope(project, resp_get_job_token_scope):
    return project.job_token_scope.get()


def test_get_job_token_scope(project, resp_get_job_token_scope):
    scope = project.job_token_scope.get()
    assert isinstance(scope, ProjectJobTokenScope)
    assert scope.inbound_enabled is True


def test_refresh_job_token_scope(job_token_scope, resp_get_job_token_scope):
    job_token_scope.refresh()
    assert job_token_scope.inbound_enabled is True


def test_save_job_token_scope(job_token_scope, resp_patch_job_token_scope):
    job_token_scope.enabled = False
    job_token_scope.save()


def test_update_job_token_scope(project, resp_patch_job_token_scope):
    project.job_token_scope.update(new_data={"enabled": False})


def test_get_projects_allowlist(job_token_scope, resp_get_allowlist):
    allowlist = job_token_scope.allowlist
    assert isinstance(allowlist, AllowlistProjectManager)

    allowlist_content = allowlist.list()
    assert isinstance(allowlist_content, list)
    assert allowlist_content[0].get_id() == 4


def test_add_project_to_allowlist(job_token_scope, resp_add_to_allowlist):
    allowlist = job_token_scope.allowlist
    assert isinstance(allowlist, AllowlistProjectManager)

    resp = allowlist.create({"target_project_id": 2})
    assert resp.get_id() == 2


def test_get_groups_allowlist(job_token_scope, resp_get_groups_allowlist):
    allowlist = job_token_scope.groups_allowlist
    assert isinstance(allowlist, AllowlistGroupManager)

    allowlist_content = allowlist.list()
    assert isinstance(allowlist_content, list)
    assert allowlist_content[0].get_id() == 4


def test_add_group_to_allowlist(job_token_scope, resp_add_to_groups_allowlist):
    allowlist = job_token_scope.groups_allowlist
    assert isinstance(allowlist, AllowlistGroupManager)

    resp = allowlist.create({"target_group_id": 4})
    assert resp.get_id() == 4



"""
GitLab API: https://docs.gitlab.com/ee/api/project_badges.html
GitLab API: https://docs.gitlab.com/ee/api/group_badges.html
"""

import re

import pytest
import responses

from gitlab.v4.objects import GroupBadge, ProjectBadge

link_url = (
    "http://example.com/ci_status.svg?project=example-org/example-project&ref=main"
)
image_url = "https://example.io/my/badge"

rendered_link_url = (
    "http://example.com/ci_status.svg?project=example-org/example-project&ref=main"
)
rendered_image_url = "https://example.io/my/badge"

new_badge = {
    "link_url": link_url,
    "image_url": image_url,
}

badge_content = {
    "name": "Coverage",
    "id": 1,
    "link_url": link_url,
    "image_url": image_url,
    "rendered_link_url": rendered_image_url,
    "rendered_image_url": rendered_image_url,
}

preview_badge_content = {
    "link_url": link_url,
    "image_url": image_url,
    "rendered_link_url": rendered_link_url,
    "rendered_image_url": rendered_image_url,
}


@pytest.fixture()
def resp_get_badge():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(r"http://localhost/api/v4/(projects|groups)/1/badges/1"),
            json=badge_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture()
def resp_list_badges():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(r"http://localhost/api/v4/(projects|groups)/1/badges"),
            json=[badge_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture()
def resp_create_badge():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url=re.compile(r"http://localhost/api/v4/(projects|groups)/1/badges"),
            json=badge_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture()
def resp_update_badge():
    updated_content = dict(badge_content)
    updated_content["link_url"] = "http://link_url"

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PUT,
            url=re.compile(r"http://localhost/api/v4/(projects|groups)/1/badges/1"),
            json=updated_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture()
def resp_delete_badge():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url=re.compile(r"http://localhost/api/v4/(projects|groups)/1/badges/1"),
            status=204,
        )
        yield rsps


@pytest.fixture()
def resp_preview_badge():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(
                r"http://localhost/api/v4/(projects|groups)/1/badges/render"
            ),
            json=preview_badge_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_project_badges(project, resp_list_badges):
    badges = project.badges.list()
    assert isinstance(badges, list)
    assert isinstance(badges[0], ProjectBadge)


def test_list_group_badges(group, resp_list_badges):
    badges = group.badges.list()
    assert isinstance(badges, list)
    assert isinstance(badges[0], GroupBadge)


def test_get_project_badge(project, resp_get_badge):
    badge = project.badges.get(1)
    assert isinstance(badge, ProjectBadge)
    assert badge.name == "Coverage"
    assert badge.id == 1


def test_get_group_badge(group, resp_get_badge):
    badge = group.badges.get(1)
    assert isinstance(badge, GroupBadge)
    assert badge.name == "Coverage"
    assert badge.id == 1


def test_delete_project_badge(project, resp_delete_badge):
    badge = project.badges.get(1, lazy=True)
    badge.delete()


def test_delete_group_badge(group, resp_delete_badge):
    badge = group.badges.get(1, lazy=True)
    badge.delete()


def test_create_project_badge(project, resp_create_badge):
    badge = project.badges.create(new_badge)
    assert isinstance(badge, ProjectBadge)
    assert badge.image_url == image_url


def test_create_group_badge(group, resp_create_badge):
    badge = group.badges.create(new_badge)
    assert isinstance(badge, GroupBadge)
    assert badge.image_url == image_url


def test_preview_project_badge(project, resp_preview_badge):
    output = project.badges.render(
        link_url=link_url,
        image_url=image_url,
    )
    assert isinstance(output, dict)
    assert "rendered_link_url" in output
    assert "rendered_image_url" in output
    assert output["link_url"] == output["rendered_link_url"]
    assert output["image_url"] == output["rendered_image_url"]


def test_preview_group_badge(group, resp_preview_badge):
    output = group.badges.render(
        link_url=link_url,
        image_url=image_url,
    )
    assert isinstance(output, dict)
    assert "rendered_link_url" in output
    assert "rendered_image_url" in output
    assert output["link_url"] == output["rendered_link_url"]
    assert output["image_url"] == output["rendered_image_url"]


def test_update_project_badge(project, resp_update_badge):
    badge = project.badges.get(1, lazy=True)
    badge.link_url = "http://link_url"
    badge.save()
    assert badge.link_url == "http://link_url"


def test_update_group_badge(group, resp_update_badge):
    badge = group.badges.get(1, lazy=True)
    badge.link_url = "http://link_url"
    badge.save()
    assert badge.link_url == "http://link_url"



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import RetrieveMixin

__all__ = [
    "Dockerfile",
    "DockerfileManager",
    "Gitignore",
    "GitignoreManager",
    "Gitlabciyml",
    "GitlabciymlManager",
    "License",
    "LicenseManager",
]


class Dockerfile(RESTObject):
    _id_attr = "name"


class DockerfileManager(RetrieveMixin, RESTManager):
    _path = "/templates/dockerfiles"
    _obj_cls = Dockerfile

    def get(self, id: Union[str, int], lazy: bool = False, **kwargs: Any) -> Dockerfile:
        return cast(Dockerfile, super().get(id=id, lazy=lazy, **kwargs))


class Gitignore(RESTObject):
    _id_attr = "name"


class GitignoreManager(RetrieveMixin, RESTManager):
    _path = "/templates/gitignores"
    _obj_cls = Gitignore

    def get(self, id: Union[str, int], lazy: bool = False, **kwargs: Any) -> Gitignore:
        return cast(Gitignore, super().get(id=id, lazy=lazy, **kwargs))


class Gitlabciyml(RESTObject):
    _id_attr = "name"


class GitlabciymlManager(RetrieveMixin, RESTManager):
    _path = "/templates/gitlab_ci_ymls"
    _obj_cls = Gitlabciyml

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> Gitlabciyml:
        return cast(Gitlabciyml, super().get(id=id, lazy=lazy, **kwargs))


class License(RESTObject):
    _id_attr = "key"


class LicenseManager(RetrieveMixin, RESTManager):
    _path = "/templates/licenses"
    _obj_cls = License
    _list_filters = ("popular",)
    _optional_get_attrs = ("project", "fullname")

    def get(self, id: Union[str, int], lazy: bool = False, **kwargs: Any) -> License:
        return cast(License, super().get(id=id, lazy=lazy, **kwargs))



import datetime


def test_project_deploy_tokens(gl, project):
    today = datetime.date.today().isoformat()
    deploy_token = project.deploytokens.create(
        {
            "name": "foo",
            "username": "bar",
            "expires_at": today,
            "scopes": ["read_registry"],
        }
    )
    assert deploy_token in project.deploytokens.list()
    assert set(project.deploytokens.list()) <= set(gl.deploytokens.list())

    deploy_token = project.deploytokens.get(deploy_token.id)
    assert deploy_token.name == "foo"
    assert deploy_token.expires_at == f"{today}T00:00:00.000Z"
    assert deploy_token.scopes == ["read_registry"]
    assert deploy_token.username == "bar"

    deploy_token.delete()


def test_group_deploy_tokens(gl, group):
    deploy_token = group.deploytokens.create(
        {
            "name": "foo",
            "scopes": ["read_registry"],
        }
    )

    assert deploy_token in group.deploytokens.list()
    assert set(group.deploytokens.list()) <= set(gl.deploytokens.list())

    deploy_token = group.deploytokens.get(deploy_token.id)
    assert deploy_token.name == "foo"
    assert deploy_token.scopes == ["read_registry"]

    deploy_token.delete()



from typing import Any, cast, Dict, Union

import requests

from gitlab import cli
from gitlab import exceptions as exc
from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import CRUDMixin, ListMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "DeployKey",
    "DeployKeyManager",
    "ProjectKey",
    "ProjectKeyManager",
]


class DeployKey(RESTObject):
    pass


class DeployKeyManager(ListMixin, RESTManager):
    _path = "/deploy_keys"
    _obj_cls = DeployKey


class ProjectKey(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class ProjectKeyManager(CRUDMixin, RESTManager):
    _path = "/projects/{project_id}/deploy_keys"
    _obj_cls = ProjectKey
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(required=("title", "key"), optional=("can_push",))
    _update_attrs = RequiredOptional(optional=("title", "can_push"))

    @cli.register_custom_action(
        cls_names="ProjectKeyManager",
        required=("key_id",),
        requires_id=False,
        help="Enable a deploy key for the project",
    )
    @exc.on_http_error(exc.GitlabProjectDeployKeyError)
    def enable(
        self, key_id: int, **kwargs: Any
    ) -> Union[Dict[str, Any], requests.Response]:
        """Enable a deploy key for a project.

        Args:
            key_id: The ID of the key to enable
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabProjectDeployKeyError: If the key could not be enabled

        Returns:
            A dict of the result.
        """
        path = f"{self.path}/{key_id}/enable"
        return self.gitlab.http_post(path, **kwargs)

    def get(self, id: Union[str, int], lazy: bool = False, **kwargs: Any) -> ProjectKey:
        return cast(ProjectKey, super().get(id=id, lazy=lazy, **kwargs))



from gitlab.mixins import (
    CreateMixin,
    CRUDMixin,
    DeleteMixin,
    GetMixin,
    ListMixin,
    NoUpdateMixin,
    RetrieveMixin,
    UpdateMixin,
)


def test_retrieve_mixin():
    class M(RetrieveMixin):
        pass

    obj = M()
    assert hasattr(obj, "list")
    assert hasattr(obj, "get")
    assert not hasattr(obj, "create")
    assert not hasattr(obj, "update")
    assert not hasattr(obj, "delete")
    assert isinstance(obj, ListMixin)
    assert isinstance(obj, GetMixin)


def test_crud_mixin():
    class M(CRUDMixin):
        pass

    obj = M()
    assert hasattr(obj, "get")
    assert hasattr(obj, "list")
    assert hasattr(obj, "create")
    assert hasattr(obj, "update")
    assert hasattr(obj, "delete")
    assert isinstance(obj, ListMixin)
    assert isinstance(obj, GetMixin)
    assert isinstance(obj, CreateMixin)
    assert isinstance(obj, UpdateMixin)
    assert isinstance(obj, DeleteMixin)


def test_no_update_mixin():
    class M(NoUpdateMixin):
        pass

    obj = M()
    assert hasattr(obj, "get")
    assert hasattr(obj, "list")
    assert hasattr(obj, "create")
    assert not hasattr(obj, "update")
    assert hasattr(obj, "delete")
    assert isinstance(obj, ListMixin)
    assert isinstance(obj, GetMixin)
    assert isinstance(obj, CreateMixin)
    assert not isinstance(obj, UpdateMixin)
    assert isinstance(obj, DeleteMixin)



release_name = "Demo Release"
release_tag_name = "v1.2.3"
release_description = "release notes go here"

link_data = {"url": "https://example.com", "name": "link_name"}


def test_create_project_release(project, project_file):
    project.refresh()  # Gets us the current default branch
    release = project.releases.create(
        {
            "name": release_name,
            "tag_name": release_tag_name,
            "description": release_description,
            "ref": project.default_branch,
        }
    )

    assert release in project.releases.list()
    assert project.releases.get(release_tag_name)
    assert release.name == release_name
    assert release.tag_name == release_tag_name
    assert release.description == release_description


def test_create_project_release_no_name(project, project_file):
    unnamed_release_tag_name = "v2.3.4"

    project.refresh()  # Gets us the current default branch
    release = project.releases.create(
        {
            "tag_name": unnamed_release_tag_name,
            "description": release_description,
            "ref": project.default_branch,
        }
    )

    assert release in project.releases.list()
    assert project.releases.get(unnamed_release_tag_name)
    assert release.tag_name == unnamed_release_tag_name
    assert release.description == release_description


def test_update_save_project_release(project, release):
    updated_description = f"{release.description} updated"
    release.description = updated_description
    release.save()

    release = project.releases.get(release.tag_name)
    assert release.description == updated_description


def test_delete_project_release(project, release):
    project.releases.delete(release.tag_name)


def test_create_project_release_links(project, release):
    release.links.create(link_data)

    release = project.releases.get(release.tag_name)
    assert release.assets["links"][0]["url"] == link_data["url"]
    assert release.assets["links"][0]["name"] == link_data["name"]



from typing import Any, Callable, cast, Dict, Iterator, Optional, TYPE_CHECKING, Union

import requests

from gitlab import cli
from gitlab import exceptions as exc
from gitlab import utils
from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import RefreshMixin, RetrieveMixin
from gitlab.types import ArrayAttribute

__all__ = [
    "ProjectJob",
    "ProjectJobManager",
]


class ProjectJob(RefreshMixin, RESTObject):
    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabJobCancelError)
    def cancel(self, **kwargs: Any) -> Dict[str, Any]:
        """Cancel the job.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabJobCancelError: If the job could not be canceled
        """
        path = f"{self.manager.path}/{self.encoded_id}/cancel"
        result = self.manager.gitlab.http_post(path, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        return result

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabJobRetryError)
    def retry(self, **kwargs: Any) -> Dict[str, Any]:
        """Retry the job.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabJobRetryError: If the job could not be retried
        """
        path = f"{self.manager.path}/{self.encoded_id}/retry"
        result = self.manager.gitlab.http_post(path, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        return result

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabJobPlayError)
    def play(self, **kwargs: Any) -> None:
        """Trigger a job explicitly.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabJobPlayError: If the job could not be triggered
        """
        path = f"{self.manager.path}/{self.encoded_id}/play"
        result = self.manager.gitlab.http_post(path, **kwargs)
        if TYPE_CHECKING:
            assert isinstance(result, dict)
        self._update_attrs(result)

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabJobEraseError)
    def erase(self, **kwargs: Any) -> None:
        """Erase the job (remove job artifacts and trace).

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabJobEraseError: If the job could not be erased
        """
        path = f"{self.manager.path}/{self.encoded_id}/erase"
        self.manager.gitlab.http_post(path, **kwargs)

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabCreateError)
    def keep_artifacts(self, **kwargs: Any) -> None:
        """Prevent artifacts from being deleted when expiration is set.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabCreateError: If the request could not be performed
        """
        path = f"{self.manager.path}/{self.encoded_id}/artifacts/keep"
        self.manager.gitlab.http_post(path, **kwargs)

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabCreateError)
    def delete_artifacts(self, **kwargs: Any) -> None:
        """Delete artifacts of a job.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabDeleteError: If the request could not be performed
        """
        path = f"{self.manager.path}/{self.encoded_id}/artifacts"
        self.manager.gitlab.http_delete(path, **kwargs)

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabGetError)
    def artifacts(
        self,
        streamed: bool = False,
        action: Optional[Callable[..., Any]] = None,
        chunk_size: int = 1024,
        *,
        iterator: bool = False,
        **kwargs: Any,
    ) -> Optional[Union[bytes, Iterator[Any]]]:
        """Get the job artifacts.

        Args:
            streamed: If True the data will be processed by chunks of
                `chunk_size` and each chunk is passed to `action` for
                treatment
            iterator: If True directly return the underlying response
                iterator
            action: Callable responsible of dealing with chunk of
                data
            chunk_size: Size of each chunk
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the artifacts could not be retrieved

        Returns:
            The artifacts if `streamed` is False, None otherwise.
        """
        path = f"{self.manager.path}/{self.encoded_id}/artifacts"
        result = self.manager.gitlab.http_get(
            path, streamed=streamed, raw=True, **kwargs
        )
        if TYPE_CHECKING:
            assert isinstance(result, requests.Response)
        return utils.response_content(
            result, streamed, action, chunk_size, iterator=iterator
        )

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabGetError)
    def artifact(
        self,
        path: str,
        streamed: bool = False,
        action: Optional[Callable[..., Any]] = None,
        chunk_size: int = 1024,
        *,
        iterator: bool = False,
        **kwargs: Any,
    ) -> Optional[Union[bytes, Iterator[Any]]]:
        """Get a single artifact file from within the job's artifacts archive.

        Args:
            path: Path of the artifact
            streamed: If True the data will be processed by chunks of
                `chunk_size` and each chunk is passed to `action` for
                treatment
            iterator: If True directly return the underlying response
                iterator
            action: Callable responsible of dealing with chunk of
                data
            chunk_size: Size of each chunk
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the artifacts could not be retrieved

        Returns:
            The artifacts if `streamed` is False, None otherwise.
        """
        path = f"{self.manager.path}/{self.encoded_id}/artifacts/{path}"
        result = self.manager.gitlab.http_get(
            path, streamed=streamed, raw=True, **kwargs
        )
        if TYPE_CHECKING:
            assert isinstance(result, requests.Response)
        return utils.response_content(
            result, streamed, action, chunk_size, iterator=iterator
        )

    @cli.register_custom_action(cls_names="ProjectJob")
    @exc.on_http_error(exc.GitlabGetError)
    def trace(
        self,
        streamed: bool = False,
        action: Optional[Callable[..., Any]] = None,
        chunk_size: int = 1024,
        *,
        iterator: bool = False,
        **kwargs: Any,
    ) -> Optional[Union[bytes, Iterator[Any]]]:
        """Get the job trace.

        Args:
            streamed: If True the data will be processed by chunks of
                `chunk_size` and each chunk is passed to `action` for
                treatment
            iterator: If True directly return the underlying response
                iterator
            action: Callable responsible of dealing with chunk of
                data
            chunk_size: Size of each chunk
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the artifacts could not be retrieved

        Returns:
            The trace
        """
        path = f"{self.manager.path}/{self.encoded_id}/trace"
        result = self.manager.gitlab.http_get(
            path, streamed=streamed, raw=True, **kwargs
        )
        if TYPE_CHECKING:
            assert isinstance(result, requests.Response)
        return utils.response_content(
            result, streamed, action, chunk_size, iterator=iterator
        )


class ProjectJobManager(RetrieveMixin, RESTManager):
    _path = "/projects/{project_id}/jobs"
    _obj_cls = ProjectJob
    _from_parent_attrs = {"project_id": "id"}
    _list_filters = ("scope",)
    _types = {"scope": ArrayAttribute}

    def get(self, id: Union[str, int], lazy: bool = False, **kwargs: Any) -> ProjectJob:
        return cast(ProjectJob, super().get(id=id, lazy=lazy, **kwargs))



"""Common mocks for resources in gitlab.v4.objects"""

import re

import pytest
import responses


@pytest.fixture
def binary_content():
    return b"binary content"


@pytest.fixture
def accepted_content():
    return {"message": "202 Accepted"}


@pytest.fixture
def created_content():
    return {"message": "201 Created"}


@pytest.fixture
def token_content():
    return {
        "user_id": 141,
        "scopes": ["api"],
        "name": "token",
        "expires_at": "2021-01-31",
        "id": 42,
        "active": True,
        "created_at": "2021-01-20T22:11:48.151Z",
        "revoked": False,
        "token": "s3cr3t",
    }


@pytest.fixture
def resp_export(accepted_content, binary_content):
    """Common fixture for group and project exports."""
    export_status_content = {
        "id": 1,
        "description": "Itaque perspiciatis minima aspernatur",
        "name": "Gitlab Test",
        "name_with_namespace": "Gitlab Org / Gitlab Test",
        "path": "gitlab-test",
        "path_with_namespace": "gitlab-org/gitlab-test",
        "created_at": "2017-08-29T04:36:44.383Z",
        "export_status": "finished",
        "_links": {
            "api_url": "https://gitlab.test/api/v4/projects/1/export/download",
            "web_url": "https://gitlab.test/gitlab-test/download_export",
        },
    }

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.POST,
            url=re.compile(r".*/api/v4/(groups|projects)/1/export"),
            json=accepted_content,
            content_type="application/json",
            status=202,
        )
        rsps.add(
            method=responses.GET,
            url=re.compile(r".*/api/v4/(groups|projects)/1/export/download"),
            body=binary_content,
            content_type="application/octet-stream",
            status=200,
        )
        # Currently only project export supports status checks
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/export",
            json=export_status_content,
            content_type="application/json",
            status=200,
        )
        yield rsps



"""
GitLab API: https://docs.gitlab.com/ee/api/jobs.html#list-pipeline-bridges
"""

import pytest
import responses

from gitlab.v4.objects import ProjectPipelineBridge


@pytest.fixture
def resp_list_bridges():
    export_bridges_content = {
        "commit": {
            "author_email": "admin@example.com",
            "author_name": "Administrator",
            "created_at": "2015-12-24T16:51:14.000+01:00",
            "id": "0ff3ae198f8601a285adcf5c0fff204ee6fba5fd",
            "message": "Test the CI integration.",
            "short_id": "0ff3ae19",
            "title": "Test the CI integration.",
        },
        "allow_failure": False,
        "created_at": "2015-12-24T15:51:21.802Z",
        "started_at": "2015-12-24T17:54:27.722Z",
        "finished_at": "2015-12-24T17:58:27.895Z",
        "duration": 240,
        "id": 7,
        "name": "teaspoon",
        "pipeline": {
            "id": 6,
            "ref": "main",
            "sha": "0ff3ae198f8601a285adcf5c0fff204ee6fba5fd",
            "status": "pending",
            "created_at": "2015-12-24T15:50:16.123Z",
            "updated_at": "2015-12-24T18:00:44.432Z",
            "web_url": "https://example.com/foo/bar/pipelines/6",
        },
        "ref": "main",
        "stage": "test",
        "status": "pending",
        "tag": False,
        "web_url": "https://example.com/foo/bar/-/jobs/7",
        "user": {
            "id": 1,
            "name": "Administrator",
            "username": "root",
            "state": "active",
            "avatar_url": "http://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80&d=identicon",
            "web_url": "http://gitlab.dev/root",
            "created_at": "2015-12-21T13:14:24.077Z",
            "public_email": "",
            "skype": "",
            "linkedin": "",
            "twitter": "",
            "website_url": "",
            "organization": "",
        },
        "downstream_pipeline": {
            "id": 5,
            "sha": "f62a4b2fb89754372a346f24659212eb8da13601",
            "ref": "main",
            "status": "pending",
            "created_at": "2015-12-24T17:54:27.722Z",
            "updated_at": "2015-12-24T17:58:27.896Z",
            "web_url": "https://example.com/diaspora/diaspora-client/pipelines/5",
        },
    }

    export_pipelines_content = [
        {
            "id": 6,
            "status": "pending",
            "ref": "new-pipeline",
            "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
            "web_url": "https://example.com/foo/bar/pipelines/47",
            "created_at": "2016-08-11T11:28:34.085Z",
            "updated_at": "2016-08-11T11:32:35.169Z",
        },
    ]

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/pipelines/6/bridges",
            json=[export_bridges_content],
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/pipelines",
            json=export_pipelines_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_projects_pipelines_bridges(project, resp_list_bridges):
    pipeline = project.pipelines.list()[0]
    bridges = pipeline.bridges.list()

    assert isinstance(bridges, list)
    assert isinstance(bridges[0], ProjectPipelineBridge)
    assert bridges[0].downstream_pipeline["id"] == 5
    assert (
        bridges[0].downstream_pipeline["sha"]
        == "f62a4b2fb89754372a346f24659212eb8da13601"
    )



from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import BadgeRenderMixin, CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "GroupBadge",
    "GroupBadgeManager",
    "ProjectBadge",
    "ProjectBadgeManager",
]


class GroupBadge(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class GroupBadgeManager(BadgeRenderMixin, CRUDMixin, RESTManager):
    _path = "/groups/{group_id}/badges"
    _obj_cls = GroupBadge
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(required=("link_url", "image_url"))
    _update_attrs = RequiredOptional(optional=("link_url", "image_url"))

    def get(self, id: Union[str, int], lazy: bool = False, **kwargs: Any) -> GroupBadge:
        return cast(GroupBadge, super().get(id=id, lazy=lazy, **kwargs))


class ProjectBadge(SaveMixin, ObjectDeleteMixin, RESTObject):
    pass


class ProjectBadgeManager(BadgeRenderMixin, CRUDMixin, RESTManager):
    _path = "/projects/{project_id}/badges"
    _obj_cls = ProjectBadge
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(required=("link_url", "image_url"))
    _update_attrs = RequiredOptional(optional=("link_url", "image_url"))

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectBadge:
        return cast(ProjectBadge, super().get(id=id, lazy=lazy, **kwargs))



import pathlib

import _pytest.config
import pytest

import gitlab


@pytest.fixture(scope="session")
def test_dir(pytestconfig: _pytest.config.Config) -> pathlib.Path:
    return pytestconfig.rootdir / "tests"  # type: ignore


@pytest.fixture(autouse=True)
def mock_clean_config(monkeypatch: pytest.MonkeyPatch) -> None:
    """Ensures user-defined environment variables do not interfere with tests."""
    monkeypatch.delenv("PYTHON_GITLAB_CFG", raising=False)
    monkeypatch.delenv("GITLAB_PRIVATE_TOKEN", raising=False)
    monkeypatch.delenv("GITLAB_URL", raising=False)
    monkeypatch.delenv("CI_JOB_TOKEN", raising=False)
    monkeypatch.delenv("CI_SERVER_URL", raising=False)


@pytest.fixture(autouse=True)
def default_files(monkeypatch: pytest.MonkeyPatch) -> None:
    """Ensures user configuration files do not interfere with tests."""
    monkeypatch.setattr(gitlab.config, "_DEFAULT_FILES", [])


@pytest.fixture
def valid_gitlab_ci_yml() -> str:
    return """---
:test_job:
  :script: echo 1
"""


@pytest.fixture
def invalid_gitlab_ci_yml() -> str:
    return "invalid"



import subprocess
import time

import pytest
import responses


@pytest.mark.script_launch_mode("inprocess")
@responses.activate
def test_project_registry_delete_in_bulk(
    script_runner, resp_delete_registry_tags_in_bulk
):
    responses.add(**resp_delete_registry_tags_in_bulk)
    cmd = [
        "gitlab",
        "project-registry-tag",
        "delete-in-bulk",
        "--project-id",
        "1",
        "--repository-id",
        "1",
        "--name-regex-delete",
        "^.*dev.*$",
        # TODO: remove `name` after deleting without ID is possible
        # See #849 and #1631
        "--name",
        ".*",
    ]
    ret = ret = script_runner.run(cmd)
    assert ret.success


@pytest.fixture
def project_export(project):
    export = project.exports.create()
    export.refresh()

    count = 0
    while export.export_status != "finished":
        time.sleep(0.5)
        export.refresh()
        count += 1
        if count == 30:
            raise Exception("Project export taking too much time")

    return export


def test_project_export_download_custom_action(gitlab_config, project_export):
    """Tests custom action on ProjectManager"""
    cmd = [
        "gitlab",
        "--config-file",
        gitlab_config,
        "project-export",
        "download",
        "--project-id",
        str(project_export.id),
    ]

    export = subprocess.run(cmd, capture_output=True, check=True)
    assert export.returncode == 0


def test_project_languages_custom_action(gitlab_cli, project, project_file):
    """Tests custom action on Project/RESTObject"""
    cmd = ["project", "languages", "--id", project.id]
    ret = gitlab_cli(cmd)
    assert ret.success



from typing import Any, List, Union

from gitlab import exceptions as exc
from gitlab.base import RESTManager, RESTObject, RESTObjectList

__all__ = [
    "LDAPGroup",
    "LDAPGroupManager",
]


class LDAPGroup(RESTObject):
    _id_attr = None


class LDAPGroupManager(RESTManager):
    _path = "/ldap/groups"
    _obj_cls = LDAPGroup
    _list_filters = ("search", "provider")

    @exc.on_http_error(exc.GitlabListError)
    def list(self, **kwargs: Any) -> Union[List[LDAPGroup], RESTObjectList]:
        """Retrieve a list of objects.

        Args:
            all: If True, return all the items, without pagination
            per_page: Number of items to retrieve per request
            page: ID of the page to return (starts with page 1)
            iterator: If set to True and no pagination option is
                defined, return a generator instead of a list
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The list of objects, or a generator if `iterator` is True

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabListError: If the server cannot perform the request
        """
        data = kwargs.copy()
        if self.gitlab.per_page:
            data.setdefault("per_page", self.gitlab.per_page)

        if "provider" in data:
            path = f"/ldap/{data['provider']}/groups"
        else:
            path = self._path

        obj = self.gitlab.http_list(path, **data)
        if isinstance(obj, list):
            return [self._obj_cls(self, item) for item in obj]
        return RESTObjectList(self, self._obj_cls, obj)



"""
GitLab API: https://docs.gitlab.com/ce/api/projects.html
"""

from unittest.mock import mock_open, patch

import pytest
import responses

from gitlab import exceptions
from gitlab.const import DEVELOPER_ACCESS, SEARCH_SCOPE_ISSUES
from gitlab.v4.objects import (
    Project,
    ProjectFork,
    ProjectUser,
    StarredProject,
    UserProject,
)
from gitlab.v4.objects.projects import ProjectStorage

project_content = {"name": "name", "id": 1}
project_with_owner_content = {
    "name": "name",
    "id": 1,
    "owner": {"id": 1, "username": "owner_username", "name": "owner_name"},
}
languages_content = {
    "python": 80.00,
    "ruby": 99.99,
    "CoffeeScript": 0.01,
}
user_content = {
    "name": "first",
    "id": 1,
    "state": "active",
}
forks_content = [
    {
        "id": 1,
    },
]
project_forked_from_content = {
    "name": "name",
    "id": 2,
    "forks_count": 0,
    "forked_from_project": {"id": 1, "name": "name", "forks_count": 1},
}
project_starrers_content = {
    "starred_since": "2019-01-28T14:47:30.642Z",
    "user": {
        "id": 1,
        "name": "name",
    },
}
upload_file_content = {
    "alt": "filename",
    "url": "/uploads/66dbcd21ec5d24ed6ea225176098d52b/filename.png",
    "full_path": "/namespace/project/uploads/66dbcd21ec5d24ed6ea225176098d52b/filename.png",
    "markdown": "![dk](/uploads/66dbcd21ec5d24ed6ea225176098d52b/filename.png)",
}
share_project_content = {
    "id": 1,
    "project_id": 1,
    "group_id": 1,
    "group_access": 30,
    "expires_at": None,
}
push_rules_content = {"id": 1, "deny_delete_tag": True}
search_issues_content = [
    {
        "id": 1,
        "iid": 1,
        "project_id": 1,
        "title": "Issue",
    }
]
pipeline_trigger_content = {
    "id": 1,
    "iid": 1,
    "project_id": 1,
    "ref": "main",
    "status": "created",
    "source": "trigger",
}


@pytest.fixture
def resp_get_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1",
            json=project_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects",
            json=project_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_create_user_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/user/1",
            json=project_with_owner_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_fork_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/fork",
            json=project_forked_from_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_update_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/projects/1",
            json=project_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_get_project_storage():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/storage",
            json={"project_id": 1, "disk_path": "/disk/path"},
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_user_projects():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/users/1/projects",
            json=[project_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_star_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/star",
            json=project_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_unstar_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/unstar",
            json=project_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_list_project_starrers():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/starrers",
            json=[project_starrers_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_starred_projects():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/users/1/starred_projects",
            json=[project_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_users():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/users",
            json=[user_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_forks():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/forks",
            json=forks_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_languages():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/languages",
            json=languages_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_projects():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects",
            json=[project_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_transfer_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/projects/1/transfer",
            json=project_content,
            content_type="application/json",
            status=200,
            match=[
                responses.matchers.json_params_matcher({"namespace": "test-namespace"})
            ],
        )
        yield rsps


@pytest.fixture
def resp_archive_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/archive",
            json=project_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_unarchive_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/unarchive",
            json=project_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_delete_project(accepted_content):
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/1",
            json=accepted_content,
            content_type="application/json",
            status=202,
        )
        yield rsps


@pytest.fixture
def resp_upload_file_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/uploads",
            json=upload_file_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_share_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/share",
            json=share_project_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_unshare_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/1/share/1",
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_create_fork_relation():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/2/fork/1",
            json=project_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_delete_fork_relation():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/2/fork",
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_trigger_pipeline():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/trigger/pipeline",
            json=pipeline_trigger_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_search_project_resources_by_name():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/search?scope=issues&search=Issue",
            json=search_issues_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_start_housekeeping():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/housekeeping",
            json={},
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_list_push_rules_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/push_rule",
            json=push_rules_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_push_rules_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/push_rule",
            json=push_rules_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_update_push_rules_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/push_rule",
            json=push_rules_content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/projects/1/push_rule",
            json=push_rules_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_delete_push_rules_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/push_rule",
            json=push_rules_content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/1/push_rule",
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_restore_project(created_content):
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/restore",
            json=created_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_start_pull_mirroring_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/mirror/pull",
            json={},
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_pull_mirror_details_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/mirror/pull",
            json={
                "id": 101486,
                "last_error": None,
                "last_successful_update_at": "2020-01-06T17:32:02.823Z",
                "last_update_at": "2020-01-06T17:32:02.823Z",
                "last_update_started_at": "2020-01-06T17:31:55.864Z",
                "update_status": "finished",
                "url": "https://*****:*****@gitlab.com/gitlab-org/security/gitlab.git",
            },
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_snapshot_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/snapshot",
            content_type="application/x-tar",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_artifact():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/jobs/artifacts/ref_name/raw/artifact_path?job=job",
            content_type="application/x-tar",
            status=200,
        )
        yield rsps


def test_get_project(gl, resp_get_project):
    data = gl.projects.get(1)
    assert isinstance(data, Project)
    assert data.name == "name"
    assert data.id == 1


def test_list_projects(gl, resp_list_projects):
    projects = gl.projects.list()
    assert isinstance(projects[0], Project)
    assert projects[0].name == "name"


def test_list_user_projects(user, resp_list_user_projects):
    user_project = user.projects.list()[0]
    assert isinstance(user_project, UserProject)
    assert user_project.name == "name"
    assert user_project.id == 1


def test_list_user_starred_projects(user, resp_list_starred_projects):
    starred_projects = user.starred_projects.list()[0]
    assert isinstance(starred_projects, StarredProject)
    assert starred_projects.name == "name"
    assert starred_projects.id == 1


def test_list_project_users(project, resp_list_users):
    user = project.users.list()[0]
    assert isinstance(user, ProjectUser)
    assert user.id == 1
    assert user.name == "first"
    assert user.state == "active"


def test_create_project(gl, resp_create_project):
    project = gl.projects.create({"name": "name"})
    assert project.id == 1
    assert project.name == "name"


def test_create_user_project(user, resp_create_user_project):
    user_project = user.projects.create({"name": "name"})
    assert user_project.id == 1
    assert user_project.name == "name"
    assert user_project.owner
    assert user_project.owner.get("id") == user.id
    assert user_project.owner.get("name") == "owner_name"
    assert user_project.owner.get("username") == "owner_username"


def test_update_project(project, resp_update_project):
    project.snippets_enabled = 1
    project.save()


def test_fork_project(project, resp_fork_project):
    fork = project.forks.create({})
    assert fork.id == 2
    assert fork.name == "name"
    assert fork.forks_count == 0
    assert fork.forked_from_project
    assert fork.forked_from_project.get("id") == project.id
    assert fork.forked_from_project.get("name") == "name"
    assert fork.forked_from_project.get("forks_count") == 1


def test_list_project_forks(project, resp_list_forks):
    fork = project.forks.list()[0]
    assert isinstance(fork, ProjectFork)
    assert fork.id == 1


def test_star_project(project, resp_star_project):
    project.star()


def test_unstar_project(project, resp_unstar_project):
    project.unstar()


@pytest.mark.skip(reason="missing test")
def test_list_project_starrers(project, resp_list_project_starrers):
    pass


def test_get_project_languages(project, resp_list_languages):
    python = project.languages().get("python")
    ruby = project.languages().get("ruby")
    coffee_script = project.languages().get("CoffeeScript")
    assert python == 80.00
    assert ruby == 99.99
    assert coffee_script == 00.01


def test_get_project_storage(project, resp_get_project_storage):
    storage = project.storage.get()
    assert isinstance(storage, ProjectStorage)
    assert storage.disk_path == "/disk/path"


def test_archive_project(project, resp_archive_project):
    project.archive()


def test_unarchive_project(project, resp_unarchive_project):
    project.unarchive()


def test_delete_project(project, resp_delete_project):
    project.delete()


def test_upload_file(project, resp_upload_file_project):
    project.upload("filename.png", "raw\nfile\ndata")


def test_upload_file_with_filepath(project, resp_upload_file_project):
    with patch("builtins.open", mock_open(read_data="raw\nfile\ndata")):
        project.upload("filename.png", None, "/filepath")


def test_upload_file_without_filepath_nor_filedata(project):
    with pytest.raises(
        exceptions.GitlabUploadError, match="No file contents or path specified"
    ):
        project.upload("filename.png")


def test_upload_file_with_filepath_and_filedata(project):
    with pytest.raises(
        exceptions.GitlabUploadError, match="File contents and file path specified"
    ):
        project.upload("filename.png", "filedata", "/filepath")


def test_share_project(project, group, resp_share_project):
    project.share(group.id, DEVELOPER_ACCESS)


def test_delete_shared_project_link(project, group, resp_unshare_project):
    project.unshare(group.id)


def test_trigger_pipeline_project(project, resp_trigger_pipeline):
    project.trigger_pipeline("MOCK_PIPELINE_TRIGGER_TOKEN", "main")


def test_create_forked_from_relationship(
    project, another_project, resp_create_fork_relation
):
    another_project.create_fork_relation(project.id)


def test_delete_forked_from_relationship(another_project, resp_delete_fork_relation):
    another_project.delete_fork_relation()


def test_search_project_resources_by_name(
    project, resp_search_project_resources_by_name
):
    issue = project.search(SEARCH_SCOPE_ISSUES, "Issue")[0]
    assert issue
    assert issue.get("title") == "Issue"


def test_project_housekeeping(project, resp_start_housekeeping):
    project.housekeeping()


def test_list_project_push_rules(project, resp_list_push_rules_project):
    pr = project.pushrules.get()
    assert pr
    assert pr.deny_delete_tag


def test_create_project_push_rule(project, resp_create_push_rules_project):
    project.pushrules.create({"deny_delete_tag": True})


def test_update_project_push_rule(
    project,
    resp_update_push_rules_project,
):
    pr = project.pushrules.get()
    pr.deny_delete_tag = False
    pr.save()


def test_delete_project_push_rule(project, resp_delete_push_rules_project):
    pr = project.pushrules.get()
    pr.delete()


def test_transfer_project(project, resp_transfer_project):
    project.transfer("test-namespace")


def test_project_pull_mirror(project, resp_start_pull_mirroring_project):
    project.mirror_pull()


def test_project_pull_mirror_details(project, resp_pull_mirror_details_project):
    details = project.mirror_pull_details()
    assert details["last_error"] is None
    assert details["update_status"] == "finished"


def test_project_restore(project, resp_restore_project):
    project.restore()


def test_project_snapshot(project, resp_snapshot_project):
    tar_file = project.snapshot()
    assert isinstance(tar_file, bytes)



import subprocess
import textwrap
import time
from io import BytesIO
from zipfile import is_zipfile

import pytest

content = textwrap.dedent(
    """\
    test-artifact:
      script: echo "test" > artifact.txt
      artifacts:
        untracked: true
    """
)
data = {
    "file_path": ".gitlab-ci.yml",
    "branch": "main",
    "content": content,
    "commit_message": "Initial commit",
}


@pytest.fixture(scope="module")
def job_with_artifacts(gitlab_runner, project):
    project.files.create(data)

    jobs = None
    while not jobs:
        time.sleep(0.5)
        jobs = project.jobs.list(scope="success")

    return project.jobs.get(jobs[0].id)


def test_cli_job_artifacts(capsysbinary, gitlab_config, job_with_artifacts):
    cmd = [
        "gitlab",
        "--config-file",
        gitlab_config,
        "project-job",
        "artifacts",
        "--id",
        str(job_with_artifacts.id),
        "--project-id",
        str(job_with_artifacts.pipeline["project_id"]),
    ]

    with capsysbinary.disabled():
        artifacts = subprocess.check_output(cmd)
    assert isinstance(artifacts, bytes)

    artifacts_zip = BytesIO(artifacts)
    assert is_zipfile(artifacts_zip)


def test_cli_project_artifact_download(gitlab_config, job_with_artifacts):
    cmd = [
        "gitlab",
        "--config-file",
        gitlab_config,
        "project-artifact",
        "download",
        "--project-id",
        str(job_with_artifacts.pipeline["project_id"]),
        "--ref-name",
        job_with_artifacts.ref,
        "--job",
        job_with_artifacts.name,
    ]

    artifacts = subprocess.run(cmd, capture_output=True, check=True)
    assert isinstance(artifacts.stdout, bytes)

    artifacts_zip = BytesIO(artifacts.stdout)
    assert is_zipfile(artifacts_zip)


def test_cli_project_artifact_raw(gitlab_config, job_with_artifacts):
    cmd = [
        "gitlab",
        "--config-file",
        gitlab_config,
        "project-artifact",
        "raw",
        "--project-id",
        str(job_with_artifacts.pipeline["project_id"]),
        "--ref-name",
        job_with_artifacts.ref,
        "--job",
        job_with_artifacts.name,
        "--artifact-path",
        "artifact.txt",
    ]

    artifacts = subprocess.run(cmd, capture_output=True, check=True)
    assert isinstance(artifacts.stdout, bytes)
    assert artifacts.stdout == b"test\n"



import subprocess
import sys
import tarfile
import zipfile
from pathlib import Path

import pytest

from gitlab._version import __title__, __version__

DOCS_DIR = "docs"
TEST_DIR = "tests"
DIST_NORMALIZED_TITLE = f"{__title__.replace('-', '_')}-{__version__}"
SDIST_FILE = f"{DIST_NORMALIZED_TITLE}.tar.gz"
WHEEL_FILE = f"{DIST_NORMALIZED_TITLE}-py{sys.version_info.major}-none-any.whl"
PY_TYPED = "gitlab/py.typed"


@pytest.fixture(scope="session")
def build(tmp_path_factory: pytest.TempPathFactory):
    temp_dir = tmp_path_factory.mktemp("build")
    subprocess.run([sys.executable, "-m", "build", "--outdir", temp_dir], check=True)
    return temp_dir


def test_sdist_includes_correct_files(build: Path) -> None:
    sdist = tarfile.open(build / SDIST_FILE, "r:gz")

    docs_dir = sdist.getmember(f"{DIST_NORMALIZED_TITLE}/{DOCS_DIR}")
    test_dir = sdist.getmember(f"{DIST_NORMALIZED_TITLE}/{TEST_DIR}")
    readme = sdist.getmember(f"{DIST_NORMALIZED_TITLE}/README.rst")
    py_typed = sdist.getmember(f"{DIST_NORMALIZED_TITLE}/{PY_TYPED}")

    assert docs_dir.isdir()
    assert test_dir.isdir()
    assert py_typed.isfile()
    assert readme.isfile()


def test_wheel_includes_correct_files(build: Path) -> None:
    wheel = zipfile.ZipFile(build / WHEEL_FILE)
    assert PY_TYPED in wheel.namelist()


def test_wheel_excludes_docs_and_tests(build: Path) -> None:
    wheel = zipfile.ZipFile(build / WHEEL_FILE)
    assert not any(file.startswith((DOCS_DIR, TEST_DIR)) for file in wheel.namelist())



import pytest

from gitlab import Gitlab
from gitlab.v4.objects import Project, ProjectRegistryProtectionRule


@pytest.fixture(scope="module", autouse=True)
def protected_registry_feature(gl: Gitlab):
    gl.features.set(name="container_registry_protected_containers", value=True)


@pytest.mark.skip(reason="Not released yet")
def test_project_protected_registry(project: Project):
    rules = project.registry_protection_rules.list()
    assert isinstance(rules, list)

    protected_registry = project.registry_protection_rules.create(
        {
            "repository_path_pattern": "test/image",
            "minimum_access_level_for_push": "maintainer",
        }
    )
    assert isinstance(protected_registry, ProjectRegistryProtectionRule)
    assert protected_registry.repository_path_pattern == "test/image"

    protected_registry.minimum_access_level_for_push = "owner"
    protected_registry.save()
    assert protected_registry.minimum_access_level_for_push == "owner"



"""
GitLab API:
https://docs.gitlab.com/ee/api/wikis.html
"""


def test_project_wikis(project):
    page = project.wikis.create({"title": "title/subtitle", "content": "test content"})
    page.content = "update content"
    page.title = "subtitle"

    page.save()

    page.delete()


def test_project_wiki_file_upload(project):
    page = project.wikis.create(
        {"title": "title/subtitle", "content": "test page content"}
    )
    filename = "test.txt"
    file_contents = "testing contents"

    uploaded_file = page.upload(filename, file_contents)

    link = uploaded_file["link"]
    file_name = uploaded_file["file_name"]
    file_path = uploaded_file["file_path"]
    assert file_name == filename
    assert file_path.startswith("uploads/")
    assert file_path.endswith(f"/{filename}")
    assert link["url"] == file_path
    assert link["markdown"] == f"[{file_name}]({file_path})"


def test_group_wikis(group):
    page = group.wikis.create({"title": "title/subtitle", "content": "test content"})
    page.content = "update content"
    page.title = "subtitle"

    page.save()

    page.delete()


def test_group_wiki_file_upload(group):
    page = group.wikis.create(
        {"title": "title/subtitle", "content": "test page content"}
    )
    filename = "test.txt"
    file_contents = "testing contents"

    uploaded_file = page.upload(filename, file_contents)

    link = uploaded_file["link"]
    file_name = uploaded_file["file_name"]
    file_path = uploaded_file["file_path"]
    assert file_name == filename
    assert file_path.startswith("uploads/")
    assert file_path.endswith(f"/{filename}")
    assert link["url"] == file_path
    assert link["markdown"] == f"[{file_name}]({file_path})"



import time

import pytest

import gitlab


@pytest.fixture
def bulk_import_enabled(gl: gitlab.Gitlab):
    settings = gl.settings.get()
    bulk_import_default = settings.bulk_import_enabled

    settings.bulk_import_enabled = True
    settings.save()

    # todo: why so fussy with feature flag timing?
    time.sleep(5)
    get_settings = gl.settings.get()
    assert get_settings.bulk_import_enabled is True

    yield settings

    settings.bulk_import_enabled = bulk_import_default
    settings.save()


# https://github.com/python-gitlab/python-gitlab/pull/2790#pullrequestreview-1873617123
@pytest.mark.xfail(reason="Bulk Imports to be worked on in a follow up")
def test_bulk_imports(gl, group, bulk_import_enabled):
    destination = f"{group.full_path}-import"
    configuration = {
        "url": gl.url,
        "access_token": gl.private_token,
    }
    migration_entity = {
        "source_full_path": group.full_path,
        "source_type": "group_entity",
        "destination_slug": destination,
        "destination_namespace": destination,
    }
    created_migration = gl.bulk_imports.create(
        {
            "configuration": configuration,
            "entities": [migration_entity],
        }
    )

    assert created_migration.source_type == "gitlab"
    assert created_migration.status == "created"

    migration = gl.bulk_imports.get(created_migration.id)
    assert migration == created_migration

    migration.refresh()
    assert migration == created_migration

    migrations = gl.bulk_imports.list()
    assert migration in migrations

    all_entities = gl.bulk_import_entities.list()
    entities = migration.entities.list()
    assert isinstance(entities, list)
    assert entities[0] in all_entities

    entity = migration.entities.get(entities[0].id)
    assert entity == entities[0]

    entity.refresh()
    assert entity.created_at == entities[0].created_at



"""
GitLab API:
https://docs.gitlab.com/ce/api/users.html
https://docs.gitlab.com/ee/api/projects.html#list-projects-starred-by-a-user
"""

import pytest
import responses

from gitlab.v4.objects import StarredProject, User, UserMembership, UserStatus

from .test_projects import project_content


@pytest.fixture
def resp_get_user():
    content = {
        "name": "name",
        "id": 1,
        "password": "password",
        "username": "username",
        "email": "email",
    }

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/users/1",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_user_memberships():
    content = [
        {
            "source_id": 1,
            "source_name": "Project one",
            "source_type": "Project",
            "access_level": "20",
        },
        {
            "source_id": 3,
            "source_name": "Group three",
            "source_type": "Namespace",
            "access_level": "20",
        },
    ]

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/users/1/memberships",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_activate():
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/users/1/activate",
            json={},
            content_type="application/json",
            status=201,
        )
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/users/1/deactivate",
            json={},
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_approve():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/users/1/approve",
            json={"message": "Success"},
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_reject():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/users/1/reject",
            json={"message": "Success"},
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_ban():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/users/1/ban",
            json={},
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_unban():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/users/1/unban",
            json={},
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_get_user_status():
    content = {
        "message": "test",
        "message_html": "<h1>Message</h1>",
        "emoji": "thumbsup",
    }

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/users/1/status",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_delete_user_identity():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/users/1/identities/test_provider",
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_follow_unfollow():
    user = {
        "id": 1,
        "username": "john_smith",
        "name": "John Smith",
        "state": "active",
        "avatar_url": "http://localhost:3000/uploads/user/avatar/1/cd8.jpeg",
        "web_url": "http://localhost:3000/john_smith",
    }
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/users/1/follow",
            json=user,
            content_type="application/json",
            status=201,
        )
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/users/1/unfollow",
            json=user,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_followers_following():
    content = [
        {
            "id": 2,
            "name": "Lennie Donnelly",
            "username": "evette.kilback",
            "state": "active",
            "avatar_url": "https://www.gravatar.com/avatar/7955171a55ac4997ed81e5976287890a?s=80&d=identicon",
            "web_url": "http://127.0.0.1:3000/evette.kilback",
        },
        {
            "id": 4,
            "name": "Serena Bradtke",
            "username": "cammy",
            "state": "active",
            "avatar_url": "https://www.gravatar.com/avatar/a2daad869a7b60d3090b7b9bef4baf57?s=80&d=identicon",
            "web_url": "http://127.0.0.1:3000/cammy",
        },
    ]
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/users/1/followers",
            json=content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/users/1/following",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_starred_projects():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/users/1/starred_projects",
            json=[project_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_runner_create():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/user/runners",
            json={"id": "6", "token": "6337ff461c94fd3fa32ba3b1ff4125"},
            content_type="application/json",
            status=201,
        )
        yield rsps


def test_get_user(gl, resp_get_user):
    user = gl.users.get(1)
    assert isinstance(user, User)
    assert user.name == "name"
    assert user.id == 1


def test_user_memberships(user, resp_get_user_memberships):
    memberships = user.memberships.list()
    assert isinstance(memberships[0], UserMembership)
    assert memberships[0].source_type == "Project"


def test_user_status(user, resp_get_user_status):
    status = user.status.get()
    assert isinstance(status, UserStatus)
    assert status.message == "test"
    assert status.emoji == "thumbsup"


def test_user_activate_deactivate(user, resp_activate):
    user.activate()
    user.deactivate()


def test_user_approve_(user, resp_approve):
    user.approve()


def test_user_approve_reject(user, resp_reject):
    user.reject()


def test_user_ban(user, resp_ban):
    user.ban()


def test_user_unban(user, resp_unban):
    user.unban()


def test_delete_user_identity(user, resp_delete_user_identity):
    user.identityproviders.delete("test_provider")


def test_user_follow_unfollow(user, resp_follow_unfollow):
    user.follow()
    user.unfollow()


def test_list_followers(user, resp_followers_following):
    followers = user.followers_users.list()
    followings = user.following_users.list()
    assert isinstance(followers[0], User)
    assert followers[0].id == 2
    assert isinstance(followings[0], User)
    assert followings[1].id == 4


def test_list_starred_projects(user, resp_starred_projects):
    projects = user.starred_projects.list()
    assert isinstance(projects[0], StarredProject)
    assert projects[0].id == project_content["id"]


def test_create_user_runner(current_user, resp_runner_create):
    runner = current_user.runners.create({"runner_type": "instance_type"})
    assert runner.id == "6"
    assert runner.token == "6337ff461c94fd3fa32ba3b1ff4125"



"""
GitLab API: https://docs.gitlab.com/ce/api/issues.html
"""

import re

import pytest
import responses

from gitlab.v4.objects import (
    GroupIssuesStatistics,
    IssuesStatistics,
    ProjectIssuesStatistics,
)


@pytest.fixture
def resp_list_issues():
    content = [{"name": "name", "id": 1}, {"name": "other_name", "id": 2}]

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/issues",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_issue():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/issues/1",
            json={"name": "name", "id": 1},
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_reorder_issue():
    match_params = {"move_after_id": 2, "move_before_id": 3}
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/projects/1/issues/1/reorder",
            json={"name": "name", "id": 1},
            content_type="application/json",
            status=200,
            match=[responses.matchers.json_params_matcher(match_params)],
        )
        yield rsps


@pytest.fixture
def resp_issue_statistics():
    content = {"statistics": {"counts": {"all": 20, "closed": 5, "opened": 15}}}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(
                r"http://localhost/api/v4/((groups|projects)/1/)?issues_statistics"
            ),
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_issues(gl, resp_list_issues):
    data = gl.issues.list()
    assert data[1].id == 2
    assert data[1].name == "other_name"


def test_get_issue(gl, resp_get_issue):
    issue = gl.issues.get(1)
    assert issue.id == 1
    assert issue.name == "name"


def test_reorder_issue(project, resp_reorder_issue):
    issue = project.issues.get(1, lazy=True)
    issue.reorder(move_after_id=2, move_before_id=3)


def test_get_issues_statistics(gl, resp_issue_statistics):
    statistics = gl.issues_statistics.get()
    assert isinstance(statistics, IssuesStatistics)
    assert statistics.statistics["counts"]["all"] == 20


def test_get_group_issues_statistics(group, resp_issue_statistics):
    statistics = group.issues_statistics.get()
    assert isinstance(statistics, GroupIssuesStatistics)
    assert statistics.statistics["counts"]["all"] == 20


def test_get_project_issues_statistics(project, resp_issue_statistics):
    statistics = project.issues_statistics.get()
    assert isinstance(statistics, ProjectIssuesStatistics)
    assert statistics.statistics["counts"]["all"] == 20



import datetime


def test_list_project_access_tokens(gitlab_cli, project):
    cmd = ["project-access-token", "list", "--project-id", project.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_create_project_access_token_with_scopes(gitlab_cli, project):
    cmd = [
        "project-access-token",
        "create",
        "--project-id",
        project.id,
        "--name",
        "test-token",
        "--scopes",
        "api,read_repository",
        "--expires-at",
        datetime.date.today().isoformat(),
    ]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_group_access_tokens(gitlab_cli, group):
    cmd = ["group-access-token", "list", "--group-id", group.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_create_group_access_token_with_scopes(gitlab_cli, group):
    cmd = [
        "group-access-token",
        "create",
        "--group-id",
        group.id,
        "--name",
        "test-token",
        "--scopes",
        "api,read_repository",
        "--expires-at",
        datetime.date.today().isoformat(),
    ]
    ret = gitlab_cli(cmd)

    assert ret.success



"""
GitLab API:
https://docs.gitlab.com/ee/api/users.html
https://docs.gitlab.com/ee/api/users.html#delete-authentication-identity-from-user
"""

import datetime
import time

import requests


def test_create_user(gl, fixture_dir):
    user = gl.users.create(
        {
            "email": "foo@bar.com",
            "username": "foo",
            "name": "foo",
            "password": "E4596f8be406Bc3a14a4ccdb1df80587$3",
            "avatar": open(fixture_dir / "avatar.png", "rb"),
        }
    )

    created_user = gl.users.list(username="foo")[0]
    assert created_user.username == user.username
    assert created_user.email == user.email

    avatar_url = user.avatar_url.replace("gitlab.test", "localhost:8080")
    uploaded_avatar = requests.get(avatar_url).content
    with open(fixture_dir / "avatar.png", "rb") as f:
        assert uploaded_avatar == f.read()


def test_block_user(gl, user):
    result = user.block()
    assert result is True
    users = gl.users.list(blocked=True)
    assert user in users

    # block again
    result = user.block()
    # Trying to block an already blocked user returns None
    assert result is None

    result = user.unblock()
    assert result is True
    users = gl.users.list(blocked=False)
    assert user in users

    # unblock again
    result = user.unblock()
    # Trying to unblock an already blocked user returns False
    assert result is False


def test_ban_user(gl, user):
    user.ban()
    retrieved_user = gl.users.get(user.id)
    assert retrieved_user.state == "banned"

    user.unban()
    retrieved_user = gl.users.get(user.id)
    assert retrieved_user.state == "active"


def test_delete_user(gl):
    new_user = gl.users.create(
        {
            "email": "delete-user@test.com",
            "username": "delete-user",
            "name": "delete-user",
            "password": "E4596f8be406Bc3a14a4ccdb1df80587#15",
        }
    )

    # We don't need to validate Gitlab's behaviour by checking if user is present after a delay etc,
    # just that python-gitlab acted correctly to produce a 2xx from Gitlab

    new_user.delete()


def test_user_projects_list(gl, user):
    projects = user.projects.list()
    assert isinstance(projects, list)
    assert not projects


def test_user_events_list(gl, user):
    events = user.events.list()
    assert isinstance(events, list)
    assert not events


def test_user_bio(gl, user):
    user.bio = "This is the user bio"
    user.save()


def test_list_multiple_users(gl, user):
    second_email = f"{user.email}.2"
    second_username = f"{user.username}_2"
    second_user = gl.users.create(
        {
            "email": second_email,
            "username": second_username,
            "name": "Foo Bar",
            "password": "E4596f8be406Bc3a14a4ccdb1df80587#!",
        }
    )
    assert gl.users.list(search=second_user.username)[0].id == second_user.id

    expected = [user, second_user]
    actual = list(gl.users.list(search=user.username))

    assert set(expected) == set(actual)
    assert not gl.users.list(search="asdf")


def test_user_gpg_keys(gl, user, GPG_KEY):
    gkey = user.gpgkeys.create({"key": GPG_KEY})
    assert gkey in user.gpgkeys.list()

    gkey.delete()


def test_user_ssh_keys(gl, user, SSH_KEY):
    key = user.keys.create({"title": "testkey", "key": SSH_KEY})
    assert key in user.keys.list()

    get_key = user.keys.get(key.id)
    assert get_key.key == key.key

    key.delete()


def test_user_email(gl, user):
    email = user.emails.create({"email": "foo2@bar.com"})
    assert email in user.emails.list()

    email.delete()


def test_user_custom_attributes(gl, user):
    user.customattributes.list()

    attr = user.customattributes.set("key", "value1")
    users_with_attribute = gl.users.list(custom_attributes={"key": "value1"})

    assert user in users_with_attribute

    assert attr.key == "key"
    assert attr.value == "value1"
    assert attr in user.customattributes.list()

    user.customattributes.set("key", "value2")
    attr_2 = user.customattributes.get("key")
    assert attr_2.value == "value2"
    assert attr_2 in user.customattributes.list()

    attr_2.delete()


def test_user_impersonation_tokens(gl, user):
    today = datetime.date.today()
    future_date = today + datetime.timedelta(days=4)

    token = user.impersonationtokens.create(
        {
            "name": "user_impersonation_token",
            "scopes": ["api", "read_user"],
            "expires_at": future_date.isoformat(),
        }
    )
    # Pause to let GL catch up (happens on hosted too, sometimes takes a while for server to be ready to merge)
    time.sleep(30)

    assert token in user.impersonationtokens.list(state="active")

    token.delete()


def test_user_identities(gl, user):
    provider = "test_provider"

    user.provider = provider
    user.extern_uid = "1"
    user.save()
    assert provider in [item["provider"] for item in user.identities]

    user.identityproviders.delete(provider)



"""
GitLab API: https://docs.gitlab.com/ee/api/merge_requests.html#list-mr-pipelines
"""

import pytest
import responses

from gitlab.v4.objects import ProjectMergeRequestPipeline

pipeline_content = {
    "id": 1,
    "sha": "959e04d7c7a30600c894bd3c0cd0e1ce7f42c11d",
    "ref": "main",
    "status": "success",
}


@pytest.fixture()
def resp_list_merge_request_pipelines():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/merge_requests/1/pipelines",
            json=[pipeline_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture()
def resp_create_merge_request_pipeline():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/merge_requests/1/pipelines",
            json=pipeline_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


def test_list_merge_requests_pipelines(project, resp_list_merge_request_pipelines):
    pipelines = project.mergerequests.get(1, lazy=True).pipelines.list()
    assert len(pipelines) == 1
    assert isinstance(pipelines[0], ProjectMergeRequestPipeline)
    assert pipelines[0].sha == pipeline_content["sha"]


def test_create_merge_requests_pipelines(project, resp_create_merge_request_pipeline):
    pipeline = project.mergerequests.get(1, lazy=True).pipelines.create()
    assert isinstance(pipeline, ProjectMergeRequestPipeline)
    assert pipeline.sha == pipeline_content["sha"]



import argparse
import contextlib
import io
import os
import sys
import tempfile
from unittest import mock

import pytest

import gitlab.base
from gitlab import cli
from gitlab.exceptions import GitlabError
from gitlab.mixins import CreateMixin, UpdateMixin
from gitlab.types import RequiredOptional
from gitlab.v4 import cli as v4_cli


@pytest.mark.parametrize(
    "gitlab_resource,expected_class",
    [
        ("class", "Class"),
        ("test-class", "TestClass"),
        ("test-longer-class", "TestLongerClass"),
        ("current-user-gpg-key", "CurrentUserGPGKey"),
        ("user-gpg-key", "UserGPGKey"),
        ("ldap-group", "LDAPGroup"),
    ],
)
def test_gitlab_resource_to_cls(gitlab_resource, expected_class):
    def _namespace():
        pass

    ExpectedClass = type(expected_class, (gitlab.base.RESTObject,), {})
    _namespace.__dict__[expected_class] = ExpectedClass

    assert cli.gitlab_resource_to_cls(gitlab_resource, _namespace) == ExpectedClass


@pytest.mark.parametrize(
    "class_name,expected_gitlab_resource",
    [
        ("Class", "class"),
        ("TestClass", "test-class"),
        ("TestUPPERCASEClass", "test-uppercase-class"),
        ("UPPERCASETestClass", "uppercase-test-class"),
        ("CurrentUserGPGKey", "current-user-gpg-key"),
        ("UserGPGKey", "user-gpg-key"),
        ("LDAPGroup", "ldap-group"),
    ],
)
def test_cls_to_gitlab_resource(class_name, expected_gitlab_resource):
    TestClass = type(class_name, (), {})

    assert cli.cls_to_gitlab_resource(TestClass) == expected_gitlab_resource


@pytest.mark.parametrize(
    "message,error,expected",
    [
        ("foobar", None, "foobar\n"),
        ("foo", GitlabError("bar"), "foo (bar)\n"),
    ],
)
def test_die(message, error, expected):
    fl = io.StringIO()
    with contextlib.redirect_stderr(fl):
        with pytest.raises(SystemExit) as test:
            cli.die(message, error)
    assert fl.getvalue() == expected
    assert test.value.code == 1


def test_parse_value():
    ret = cli._parse_value("foobar")
    assert ret == "foobar"

    ret = cli._parse_value(True)
    assert ret is True

    ret = cli._parse_value(1)
    assert ret == 1

    ret = cli._parse_value(None)
    assert ret is None

    fd, temp_path = tempfile.mkstemp()
    os.write(fd, b"content")
    os.close(fd)
    ret = cli._parse_value(f"@{temp_path}")
    assert ret == "content"
    os.unlink(temp_path)

    fl = io.StringIO()
    with contextlib.redirect_stderr(fl):
        with pytest.raises(SystemExit) as exc:
            cli._parse_value("@/thisfileprobablydoesntexist")
        assert fl.getvalue().startswith(
            "FileNotFoundError: [Errno 2] No such file or directory:"
        )
        assert exc.value.code == 1


def test_base_parser():
    parser = cli._get_base_parser()
    args = parser.parse_args(["-v", "-g", "gl_id", "-c", "foo.cfg", "-c", "bar.cfg"])
    assert args.verbose
    assert args.gitlab == "gl_id"
    assert args.config_file == ["foo.cfg", "bar.cfg"]
    assert args.ssl_verify is None


def test_no_ssl_verify():
    parser = cli._get_base_parser()
    args = parser.parse_args(["--no-ssl-verify"])
    assert args.ssl_verify is False


def test_v4_parse_args():
    parser = cli._get_parser()
    args = parser.parse_args(["project", "list"])
    assert args.gitlab_resource == "project"
    assert args.resource_action == "list"


def test_v4_parser():
    parser = cli._get_parser()
    subparsers = next(
        action
        for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)
    )
    assert subparsers is not None
    assert "project" in subparsers.choices

    user_subparsers = next(
        action
        for action in subparsers.choices["project"]._actions
        if isinstance(action, argparse._SubParsersAction)
    )
    assert user_subparsers is not None
    assert "list" in user_subparsers.choices
    assert "get" in user_subparsers.choices
    assert "delete" in user_subparsers.choices
    assert "update" in user_subparsers.choices
    assert "create" in user_subparsers.choices
    assert "archive" in user_subparsers.choices
    assert "unarchive" in user_subparsers.choices

    actions = user_subparsers.choices["create"]._option_string_actions
    assert not actions["--description"].required

    user_subparsers = next(
        action
        for action in subparsers.choices["group"]._actions
        if isinstance(action, argparse._SubParsersAction)
    )
    actions = user_subparsers.choices["create"]._option_string_actions
    assert actions["--name"].required


def test_extend_parser():
    class ExceptionArgParser(argparse.ArgumentParser):
        def error(self, message):
            "Raise error instead of exiting on invalid arguments, to make testing easier"
            raise ValueError(message)

    class Fake:
        _id_attr = None

    class FakeManager(gitlab.base.RESTManager, CreateMixin, UpdateMixin):
        _obj_cls = Fake
        _create_attrs = RequiredOptional(
            required=("create",),
            optional=("opt_create",),
            exclusive=("create_a", "create_b"),
        )
        _update_attrs = RequiredOptional(
            required=("update",),
            optional=("opt_update",),
            exclusive=("update_a", "update_b"),
        )

    parser = ExceptionArgParser()
    with mock.patch.dict(
        "gitlab.v4.objects.__dict__", {"FakeManager": FakeManager}, clear=True
    ):
        v4_cli.extend_parser(parser)

    assert parser.parse_args(["fake", "create", "--create", "1"])
    assert parser.parse_args(["fake", "create", "--create", "1", "--opt-create", "1"])
    assert parser.parse_args(["fake", "create", "--create", "1", "--create-a", "1"])
    assert parser.parse_args(["fake", "create", "--create", "1", "--create-b", "1"])

    with pytest.raises(ValueError):
        # missing required "create"
        parser.parse_args(["fake", "create", "--opt_create", "1"])

    with pytest.raises(ValueError):
        # both exclusive options
        parser.parse_args(
            ["fake", "create", "--create", "1", "--create-a", "1", "--create-b", "1"]
        )

    assert parser.parse_args(["fake", "update", "--update", "1"])
    assert parser.parse_args(["fake", "update", "--update", "1", "--opt-update", "1"])
    assert parser.parse_args(["fake", "update", "--update", "1", "--update-a", "1"])
    assert parser.parse_args(["fake", "update", "--update", "1", "--update-b", "1"])

    with pytest.raises(ValueError):
        # missing required "update"
        parser.parse_args(["fake", "update", "--opt_update", "1"])

    with pytest.raises(ValueError):
        # both exclusive options
        parser.parse_args(
            ["fake", "update", "--update", "1", "--update-a", "1", "--update-b", "1"]
        )


@pytest.mark.skipif(sys.version_info < (3, 8), reason="added in 3.8")
def test_legacy_display_without_fields_warns(fake_object_no_id):
    printer = v4_cli.LegacyPrinter()

    with mock.patch("builtins.print") as mocked:
        printer.display(fake_object_no_id, obj=fake_object_no_id)

    assert "No default fields to show" in mocked.call_args.args[0]


@pytest.mark.skipif(sys.version_info < (3, 8), reason="added in 3.8")
def test_legacy_display_with_long_repr_truncates(fake_object_long_repr):
    printer = v4_cli.LegacyPrinter()

    with mock.patch("builtins.print") as mocked:
        printer.display(fake_object_long_repr, obj=fake_object_long_repr)

    assert len(mocked.call_args.args[0]) < 80



"""
GitLab API:
https://docs.gitlab.com/ee/api/integrations.html
"""

import gitlab


def test_get_service_lazy(project):
    service = project.services.get("jira", lazy=True)
    assert isinstance(service, gitlab.v4.objects.ProjectService)


def test_update_service(project):
    service_dict = project.services.update(
        "emails-on-push", {"recipients": "email@example.com"}
    )
    assert service_dict["active"]


def test_list_services(project, service):
    services = project.services.list()
    assert isinstance(services[0], gitlab.v4.objects.ProjectService)
    assert services[0].active


def test_get_service(project, service):
    service_object = project.services.get(service["slug"])
    assert isinstance(service_object, gitlab.v4.objects.ProjectService)
    assert service_object.active


def test_delete_service(project, service):
    service_object = project.services.get(service["slug"])

    service_object.delete()



"""
GitLab API:
https://docs.gitlab.com/ee/api/instance_level_ci_variables.html
https://docs.gitlab.com/ee/api/project_level_variables.html
https://docs.gitlab.com/ee/api/group_level_variables.html
"""

from typing import Any, cast, Union

from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import CRUDMixin, ObjectDeleteMixin, SaveMixin
from gitlab.types import RequiredOptional

__all__ = [
    "Variable",
    "VariableManager",
    "GroupVariable",
    "GroupVariableManager",
    "ProjectVariable",
    "ProjectVariableManager",
]


class Variable(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class VariableManager(CRUDMixin, RESTManager):
    _path = "/admin/ci/variables"
    _obj_cls = Variable
    _create_attrs = RequiredOptional(
        required=("key", "value"), optional=("protected", "variable_type", "masked")
    )
    _update_attrs = RequiredOptional(
        required=("key", "value"), optional=("protected", "variable_type", "masked")
    )

    def get(self, id: Union[str, int], lazy: bool = False, **kwargs: Any) -> Variable:
        return cast(Variable, super().get(id=id, lazy=lazy, **kwargs))


class GroupVariable(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class GroupVariableManager(CRUDMixin, RESTManager):
    _path = "/groups/{group_id}/variables"
    _obj_cls = GroupVariable
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = RequiredOptional(
        required=("key", "value"), optional=("protected", "variable_type", "masked")
    )
    _update_attrs = RequiredOptional(
        required=("key", "value"), optional=("protected", "variable_type", "masked")
    )

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> GroupVariable:
        return cast(GroupVariable, super().get(id=id, lazy=lazy, **kwargs))


class ProjectVariable(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class ProjectVariableManager(CRUDMixin, RESTManager):
    _path = "/projects/{project_id}/variables"
    _obj_cls = ProjectVariable
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = RequiredOptional(
        required=("key", "value"),
        optional=("protected", "variable_type", "masked", "environment_scope"),
    )
    _update_attrs = RequiredOptional(
        required=("key", "value"),
        optional=("protected", "variable_type", "masked", "environment_scope"),
    )

    def get(
        self, id: Union[str, int], lazy: bool = False, **kwargs: Any
    ) -> ProjectVariable:
        return cast(ProjectVariable, super().get(id=id, lazy=lazy, **kwargs))



"""
GitLab API: https://docs.gitlab.com/ce/api/packages.html
"""

import re

import pytest
import responses

from gitlab import exceptions as exc
from gitlab.v4.objects import (
    GenericPackage,
    GroupPackage,
    ProjectPackage,
    ProjectPackageFile,
    ProjectPackagePipeline,
)

package_content = {
    "id": 1,
    "name": "com/mycompany/my-app",
    "version": "1.0-SNAPSHOT",
    "package_type": "maven",
    "_links": {
        "web_path": "/namespace1/project1/-/packages/1",
        "delete_api_path": "/namespace1/project1/-/packages/1",
    },
    "created_at": "2019-11-27T03:37:38.711Z",
    "pipeline": {
        "id": 123,
        "status": "pending",
        "ref": "new-pipeline",
        "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
        "web_url": "https://example.com/foo/bar/pipelines/47",
        "created_at": "2016-08-11T11:28:34.085Z",
        "updated_at": "2016-08-11T11:32:35.169Z",
        "user": {
            "name": "Administrator",
            "avatar_url": "https://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80&d=identicon",
        },
    },
    "versions": [
        {
            "id": 2,
            "version": "2.0-SNAPSHOT",
            "created_at": "2020-04-28T04:42:11.573Z",
            "pipeline": {
                "id": 234,
                "status": "pending",
                "ref": "new-pipeline",
                "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
                "web_url": "https://example.com/foo/bar/pipelines/58",
                "created_at": "2016-08-11T11:28:34.085Z",
                "updated_at": "2016-08-11T11:32:35.169Z",
                "user": {
                    "name": "Administrator",
                    "avatar_url": "https://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80&d=identicon",
                },
            },
        }
    ],
}

package_file_content = [
    {
        "id": 25,
        "package_id": 1,
        "created_at": "2018-11-07T15:25:52.199Z",
        "file_name": "my-app-1.5-20181107.152550-1.jar",
        "size": 2421,
        "file_md5": "58e6a45a629910c6ff99145a688971ac",
        "file_sha1": "ebd193463d3915d7e22219f52740056dfd26cbfe",
        "pipelines": [
            {
                "id": 123,
                "status": "pending",
                "ref": "new-pipeline",
                "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
                "web_url": "https://example.com/foo/bar/pipelines/47",
                "created_at": "2016-08-11T11:28:34.085Z",
                "updated_at": "2016-08-11T11:32:35.169Z",
                "user": {
                    "name": "Administrator",
                    "avatar_url": "https://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80&d=identicon",
                },
            }
        ],
    },
    {
        "id": 26,
        "package_id": 1,
        "created_at": "2018-11-07T15:25:56.776Z",
        "file_name": "my-app-1.5-20181107.152550-1.pom",
        "size": 1122,
        "file_md5": "d90f11d851e17c5513586b4a7e98f1b2",
        "file_sha1": "9608d068fe88aff85781811a42f32d97feb440b5",
    },
    {
        "id": 27,
        "package_id": 1,
        "created_at": "2018-11-07T15:26:00.556Z",
        "file_name": "maven-metadata.xml",
        "size": 767,
        "file_md5": "6dfd0cce1203145a927fef5e3a1c650c",
        "file_sha1": "d25932de56052d320a8ac156f745ece73f6a8cd2",
    },
]

package_pipeline_content = [
    {
        "id": 123,
        "iid": 1,
        "project_id": 1,
        "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
        "ref": "new-pipeline",
        "status": "failed",
        "source": "push",
        "created_at": "2016-08-11T11:28:34.085Z",
        "updated_at": "2016-08-11T11:32:35.169Z",
        "web_url": "https://example.com/foo/bar/pipelines/47",
        "user": {
            "id": 1,
            "username": "root",
            "name": "Administrator",
            "state": "active",
            "avatar_url": "https://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80\u0026d=identicon",
            "web_url": "http://gdk.test:3001/root",
        },
    },
    {
        "id": 234,
        "iid": 2,
        "project_id": 1,
        "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
        "ref": "new-pipeline",
        "status": "failed",
        "source": "push",
        "created_at": "2016-08-11T11:28:34.085Z",
        "updated_at": "2016-08-11T11:32:35.169Z",
        "web_url": "https://example.com/foo/bar/pipelines/58",
        "user": {
            "id": 1,
            "username": "root",
            "name": "Administrator",
            "state": "active",
            "avatar_url": "https://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80\u0026d=identicon",
            "web_url": "http://gdk.test:3001/root",
        },
    },
]


package_name = "hello-world"
package_version = "v1.0.0"
file_name = "hello.tar.gz"
file_content = "package content"
package_url = f"http://localhost/api/v4/projects/1/packages/generic/{package_name}/{package_version}/{file_name}"


@pytest.fixture
def resp_list_packages():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(r"http://localhost/api/v4/(groups|projects)/1/packages"),
            json=[package_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_package():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/packages/1",
            json=package_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_delete_package():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/1/packages/1",
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_delete_package_file():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/1/packages/1/package_files/1",
            status=204,
        )
        yield rsps


@pytest.fixture
def resp_delete_package_file_list():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(
                r"http://localhost/api/v4/projects/1/packages/1/package_files"
            ),
            json=package_file_content,
            content_type="application/json",
            status=200,
        )
        for pkg_file_id in range(25, 28):
            rsps.add(
                method=responses.DELETE,
                url=f"http://localhost/api/v4/projects/1/packages/1/package_files/{pkg_file_id}",
                status=204,
            )
        yield rsps


@pytest.fixture
def resp_list_package_files():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(
                r"http://localhost/api/v4/projects/1/packages/1/package_files"
            ),
            json=package_file_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_package_pipelines():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(r"http://localhost/api/v4/projects/1/packages/1/pipelines"),
            json=package_pipeline_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_upload_generic_package(created_content):
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.PUT,
            url=package_url,
            json=created_content,
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_download_generic_package(created_content):
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=package_url,
            body=file_content,
            content_type="application/octet-stream",
            status=200,
        )
        yield rsps


def test_list_project_packages(project, resp_list_packages):
    packages = project.packages.list()
    assert isinstance(packages, list)
    assert isinstance(packages[0], ProjectPackage)
    assert packages[0].version == "1.0-SNAPSHOT"


def test_list_group_packages(group, resp_list_packages):
    packages = group.packages.list()
    assert isinstance(packages, list)
    assert isinstance(packages[0], GroupPackage)
    assert packages[0].version == "1.0-SNAPSHOT"


def test_get_project_package(project, resp_get_package):
    package = project.packages.get(1)
    assert isinstance(package, ProjectPackage)
    assert package.version == "1.0-SNAPSHOT"


def test_delete_project_package(project, resp_delete_package):
    package = project.packages.get(1, lazy=True)
    package.delete()


def test_list_project_package_files(project, resp_list_package_files):
    package = project.packages.get(1, lazy=True)
    package_files = package.package_files.list()
    assert isinstance(package_files, list)
    assert isinstance(package_files[0], ProjectPackageFile)
    assert package_files[0].id == 25


def test_delete_project_package_file_from_package_object(
    project, resp_delete_package_file
):
    package = project.packages.get(1, lazy=True)
    package.package_files.delete(1)


def test_delete_project_package_file_from_package_file_object(
    project, resp_delete_package_file_list
):
    package = project.packages.get(1, lazy=True)
    for package_file in package.package_files.list():
        package_file.delete()


def test_list_project_package_pipelines(project, resp_list_package_pipelines):
    package = project.packages.get(1, lazy=True)
    pipelines = package.pipelines.list()
    assert isinstance(pipelines, list)
    assert isinstance(pipelines[0], ProjectPackagePipeline)
    assert pipelines[0].id == 123


def test_upload_generic_package(tmp_path, project, resp_upload_generic_package):
    path = tmp_path / file_name
    path.write_text(file_content, encoding="utf-8")
    package = project.generic_packages.upload(
        package_name=package_name,
        package_version=package_version,
        file_name=file_name,
        path=path,
    )

    assert isinstance(package, GenericPackage)


def test_upload_generic_package_nonexistent_path(tmp_path, project):
    with pytest.raises(exc.GitlabUploadError):
        project.generic_packages.upload(
            package_name=package_name,
            package_version=package_version,
            file_name=file_name,
            path="bad",
        )


def test_upload_generic_package_no_file_and_no_data(tmp_path, project):
    path = tmp_path / file_name

    path.write_text(file_content, encoding="utf-8")

    with pytest.raises(exc.GitlabUploadError):
        project.generic_packages.upload(
            package_name=package_name,
            package_version=package_version,
            file_name=file_name,
        )


def test_upload_generic_package_file_and_data(tmp_path, project):
    path = tmp_path / file_name

    path.write_text(file_content, encoding="utf-8")

    with pytest.raises(exc.GitlabUploadError):
        project.generic_packages.upload(
            package_name=package_name,
            package_version=package_version,
            file_name=file_name,
            path=path,
            data=path.read_bytes(),
        )


def test_upload_generic_package_bytes(tmp_path, project, resp_upload_generic_package):
    path = tmp_path / file_name

    path.write_text(file_content, encoding="utf-8")

    package = project.generic_packages.upload(
        package_name=package_name,
        package_version=package_version,
        file_name=file_name,
        data=path.read_bytes(),
    )

    assert isinstance(package, GenericPackage)


def test_upload_generic_package_file(tmp_path, project, resp_upload_generic_package):
    path = tmp_path / file_name

    path.write_text(file_content, encoding="utf-8")

    package = project.generic_packages.upload(
        package_name=package_name,
        package_version=package_version,
        file_name=file_name,
        data=path.open(mode="rb"),
    )

    assert isinstance(package, GenericPackage)


def test_download_generic_package(project, resp_download_generic_package):
    package = project.generic_packages.download(
        package_name=package_name,
        package_version=package_version,
        file_name=file_name,
    )

    assert isinstance(package, bytes)



"""
GitLab API: https://docs.gitlab.com/ce/api/todos.html
"""

import pytest
import responses

from gitlab.v4.objects import Todo


@pytest.fixture
def json_content():
    return [
        {
            "id": 102,
            "project": {
                "id": 2,
                "name": "Gitlab Ce",
                "name_with_namespace": "Gitlab Org / Gitlab Ce",
                "path": "gitlab-ce",
                "path_with_namespace": "gitlab-org/gitlab-ce",
            },
            "author": {
                "name": "Administrator",
                "username": "root",
                "id": 1,
            },
            "action_name": "marked",
            "target_type": "MergeRequest",
            "target": {
                "id": 34,
                "iid": 7,
                "project_id": 2,
                "assignee": {
                    "name": "Administrator",
                    "username": "root",
                    "id": 1,
                },
            },
        }
    ]


@pytest.fixture
def resp_todo(json_content):
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/todos",
            json=json_content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/todos/102/mark_as_done",
            json=json_content[0],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_mark_all_as_done():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/todos/mark_as_done",
            status=204,
        )
        yield rsps


def test_todo(gl, resp_todo):
    todo = gl.todos.list()[0]
    assert isinstance(todo, Todo)
    assert todo.id == 102
    assert todo.target_type == "MergeRequest"
    assert todo.target["assignee"]["username"] == "root"

    todo.mark_as_done()


def test_todo_mark_all_as_done(gl, resp_mark_all_as_done):
    gl.todos.mark_all_as_done()



from gitlab.base import RESTManager, RESTObject
from gitlab.mixins import ListMixin

__all__ = [
    "ProjectMergeTrain",
    "ProjectMergeTrainManager",
]


class ProjectMergeTrain(RESTObject):
    pass


class ProjectMergeTrainManager(ListMixin, RESTManager):
    _path = "/projects/{project_id}/merge_trains"
    _obj_cls = ProjectMergeTrain
    _from_parent_attrs = {"project_id": "id"}
    _list_filters = ("scope",)



"""
pytest-docker fixture overrides.
See https://github.com/avast/pytest-docker#available-fixtures.
"""

import pytest


@pytest.fixture(scope="session")
def docker_compose_project_name():
    """Set a consistent project name to enable optional reuse of containers."""
    return "pytest-python-gitlab"


@pytest.fixture(scope="session")
def docker_compose_file(fixture_dir):
    return fixture_dir / "docker-compose.yml"


@pytest.fixture(scope="session")
def docker_cleanup(request):
    """Conditionally keep containers around by overriding the cleanup command."""
    if request.config.getoption("--keep-containers"):
        # Print version and exit.
        return "-v"
    return "down -v"



import pytest
import responses

import gitlab
from tests.unit import helpers


@pytest.fixture
def fake_manager(gl):
    return helpers.FakeManager(gl)


@pytest.fixture
def fake_manager_with_parent(gl, fake_manager):
    return helpers.FakeManagerWithParent(
        gl, parent=helpers.FakeParent(manager=fake_manager, attrs={})
    )


@pytest.fixture
def fake_object(fake_manager):
    return helpers.FakeObject(fake_manager, {"attr1": "foo", "alist": [1, 2, 3]})


@pytest.fixture
def fake_object_no_id(fake_manager):
    return helpers.FakeObjectWithoutId(fake_manager, {})


@pytest.fixture
def fake_object_long_repr(fake_manager):
    return helpers.FakeObjectWithLongRepr(fake_manager, {"test": "a" * 100})


@pytest.fixture
def fake_object_with_parent(fake_manager_with_parent):
    return helpers.FakeObject(
        fake_manager_with_parent, {"attr1": "foo", "alist": [1, 2, 3]}
    )


@pytest.fixture
def gl():
    return gitlab.Gitlab(
        "http://localhost",
        private_token="private_token",
        ssl_verify=True,
        api_version="4",
    )


@pytest.fixture
def gl_retry():
    return gitlab.Gitlab(
        "http://localhost",
        private_token="private_token",
        ssl_verify=True,
        api_version="4",
        retry_transient_errors=True,
    )


@pytest.fixture
def resp_get_current_user():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/user",
            json={
                "id": 1,
                "username": "username",
                "web_url": "http://localhost/username",
            },
            content_type="application/json",
            status=200,
        )
        yield rsps


# Todo: parametrize, but check what tests it's really useful for
@pytest.fixture
def gl_trailing():
    return gitlab.Gitlab(
        "http://localhost/", private_token="private_token", api_version="4"
    )


@pytest.fixture
def default_config(tmpdir):
    valid_config = """[global]
    default = one
    ssl_verify = true
    timeout = 2

    [one]
    url = http://one.url
    private_token = ABCDEF
    """

    config_path = tmpdir.join("python-gitlab.cfg")
    config_path.write(valid_config)
    return str(config_path)


@pytest.fixture
def tag_name():
    return "v1.0.0"


@pytest.fixture
def group(gl):
    return gl.groups.get(1, lazy=True)


@pytest.fixture
def project(gl):
    return gl.projects.get(1, lazy=True)


@pytest.fixture
def another_project(gl):
    return gl.projects.get(2, lazy=True)


@pytest.fixture
def project_issue(project):
    return project.issues.get(1, lazy=True)


@pytest.fixture
def project_merge_request(project):
    return project.mergerequests.get(1, lazy=True)


@pytest.fixture
def release(project, tag_name):
    return project.releases.get(tag_name, lazy=True)


@pytest.fixture
def schedule(project):
    return project.pipelineschedules.get(1, lazy=True)


@pytest.fixture
def user(gl):
    return gl.users.get(1, lazy=True)


@pytest.fixture
def current_user(gl, resp_get_current_user):
    gl.auth()
    return gl.user


@pytest.fixture
def migration(gl):
    return gl.bulk_imports.get(1, lazy=True)



import time

import pytest

import gitlab


# https://github.com/python-gitlab/python-gitlab/pull/2790#pullrequestreview-1873617123
def test_group_import_export(gl, group, temp_dir):
    export = group.exports.create()
    assert export.message == "202 Accepted"

    # We cannot check for export_status with group export API
    time.sleep(10)

    import_archive = temp_dir / "gitlab-group-export.tgz"
    import_path = "imported_group"
    import_name = "Imported Group"

    with open(import_archive, "wb") as f:
        export.download(streamed=True, action=f.write)

    with open(import_archive, "rb") as f:
        output = gl.groups.import_group(f, import_path, import_name)
    assert output["message"] == "202 Accepted"

    # We cannot check for returned ID with group import API
    time.sleep(10)
    group_import = gl.groups.get(import_path)

    assert group_import.path == import_path
    assert group_import.name == import_name


# https://github.com/python-gitlab/python-gitlab/pull/2790#pullrequestreview-1873617123
@pytest.mark.xfail(reason="test_project_import_export to be worked on in a follow up")
def test_project_import_export(gl, project, temp_dir):
    export = project.exports.create()
    assert export.message == "202 Accepted"

    export = project.exports.get()
    assert isinstance(export, gitlab.v4.objects.ProjectExport)

    count = 0
    while export.export_status != "finished":
        time.sleep(1)
        export.refresh()
        count += 1
        if count == 15:
            raise Exception("Project export taking too much time")

    with open(temp_dir / "gitlab-export.tgz", "wb") as f:
        export.download(streamed=True, action=f.write)  # type: ignore[arg-type]

    output = gl.projects.import_project(
        open(temp_dir / "gitlab-export.tgz", "rb"),
        "imported_project",
        name="Imported Project",
    )
    project_import = gl.projects.get(output["id"], lazy=True).imports.get()

    assert project_import.path == "imported_project"
    assert project_import.name == "Imported Project"

    count = 0
    while project_import.import_status != "finished":
        time.sleep(1)
        project_import.refresh()
        count += 1
        if count == 15:
            raise Exception("Project import taking too much time")


# https://github.com/python-gitlab/python-gitlab/pull/2790#pullrequestreview-1873617123
@pytest.mark.xfail(reason="test_project_remote_import to be worked on in a follow up")
def test_project_remote_import(gl):
    with pytest.raises(gitlab.exceptions.GitlabImportError) as err_info:
        gl.projects.remote_import(
            "ftp://whatever.com/url", "remote-project", "remote-project", "root"
        )
    assert err_info.value.response_code == 400
    assert (
        "File url is blocked: Only allowed schemes are https"
        in err_info.value.error_message
    )


# https://github.com/python-gitlab/python-gitlab/pull/2790#pullrequestreview-1873617123
@pytest.mark.xfail(
    reason="test_project_remote_import_s3 to be worked on in a follow up"
)
def test_project_remote_import_s3(gl):
    gl.features.set("import_project_from_remote_file_s3", True)
    with pytest.raises(gitlab.exceptions.GitlabImportError) as err_info:
        gl.projects.remote_import_s3(
            "remote-project",
            "aws-region",
            "aws-bucket-name",
            "aws-file-key",
            "aws-access-key-id",
            "secret-access-key",
            "remote-project",
            "root",
        )
    assert err_info.value.response_code == 400
    assert (
        "Failed to open 'aws-file-key' in 'aws-bucket-name'"
        in err_info.value.error_message
    )
